import weakref

import numpy as np
import pytest

import asdf


def test_asdffile_tree_cleared_on_close(tmp_path):
    """
    closed AsdfFile instances hold private reference to tree

    https://github.com/asdf-format/asdf/issues/1558
    """

    fn = tmp_path / "test.asdf"
    asdf.AsdfFile({"a": np.arange(1000), "b": np.arange(42)}).write_to(fn)

    with asdf.open(fn, memmap=False, lazy_load=False) as af:
        array_weakref = weakref.ref(af["a"])
        array_ref = af["b"]

    # we shouldn't be able to access the now closed file
    with pytest.raises(OSError, match="Cannot access data"):
        af["a"]
    with pytest.raises(OSError, match="Cannot access data"):
        af["b"]

    # we also should not be able to resolve the weak reference
    # meaning the tree should have been cleaned up
    r = array_weakref()
    assert r is None

    assert array_ref is not None



import numpy as np

import asdf
from asdf.versioning import supported_versions


def generate_file(path, version):
    if version not in supported_versions:
        msg = f"ASDF Standard version {version} is not supported by version {asdf.__version__} of the asdf library"
        raise ValueError(msg)

    af = asdf.AsdfFile({"array": np.ones((8, 16))}, version=version)
    af.write_to(path)


def assert_file_correct(path):
    __tracebackhide__ = True

    with asdf.open(str(path)) as af:
        assert af["array"].shape == (8, 16)
        assert np.all(af["array"] == 1)



import os
import sys

import numpy as np
import pytest

import asdf


@pytest.mark.skipif(
    sys.platform.startswith("win"), reason="os.pipe.seek noop on windows: https://bugs.python.org/issue42602"
)
def test_failure_to_write_blocks_to_non_seekable_file():
    """
    ASDF fails to write blocks to non-seekable file

    https://github.com/asdf-format/asdf/issues/1542
    """
    r, w = os.pipe()
    with os.fdopen(r, "rb") as rf:
        with os.fdopen(w, "wb") as wf:
            arrs = [np.zeros(1, dtype="uint8") + i for i in range(10)]
            af = asdf.AsdfFile({"arrs": arrs})
            af.write_to(wf)
        with asdf.open(rf) as raf:
            for a, ra in zip(arrs, raf["arrs"]):
                np.testing.assert_array_equal(a, ra)



import copy
import datetime
import json
import warnings
from collections import OrderedDict
from collections.abc import Mapping
from functools import lru_cache
from numbers import Integral
from operator import methodcaller

import numpy as np
import yaml

from asdf._jsonschema import validators as mvalidators
from asdf._jsonschema.exceptions import RefResolutionError, ValidationError

from . import constants, generic_io, reference, tagged, treeutil, util, versioning, yamlutil
from .config import get_config
from .exceptions import AsdfWarning
from .util import _patched_urllib_parse

YAML_SCHEMA_METASCHEMA_ID = "http://stsci.edu/schemas/yaml-schema/draft-01"

__all__ = ["check_schema", "fill_defaults", "load_schema", "remove_defaults", "validate"]

PYTHON_TYPE_TO_YAML_TAG = {
    None: "null",
    str: "str",
    bytes: "str",
    bool: "bool",
    int: "int",
    float: "float",
    list: "seq",
    dict: "map",
    set: "set",
    OrderedDict: "omap",
}


# Prepend full YAML tag prefix
for k, v in PYTHON_TYPE_TO_YAML_TAG.items():
    PYTHON_TYPE_TO_YAML_TAG[k] = constants.YAML_TAG_PREFIX + v


def _type_to_tag(type_):
    for base in type_.mro():
        if base in PYTHON_TYPE_TO_YAML_TAG:
            return PYTHON_TYPE_TO_YAML_TAG[base]
    return None


def validate_tag(validator, tag_pattern, instance, schema):
    """
    Implements the tag validation directive, which checks the
    tag against a pattern that may include wildcards.  See
    `asdf.util.uri_match` for details on the matching behavior.
    """
    instance_tag = instance._tag if hasattr(instance, "_tag") else _type_to_tag(type(instance))

    if instance_tag is None:
        yield ValidationError(
            f"mismatched tags, wanted '{tag_pattern}', got unhandled object type '{util.get_class_name(instance)}'",
        )

    if not util.uri_match(tag_pattern, instance_tag):
        yield ValidationError(f"mismatched tags, wanted '{tag_pattern}', got '{instance_tag}'")


def validate_propertyOrder(validator, order, instance, schema):
    """
    Stores a value on the `tagged.TaggedDict` instance so that
    properties can be written out in the preferred order.  In that
    sense this isn't really a "validator", but using the `jsonschema`
    library's extensible validation system is the easiest way to get
    this property assigned.
    """
    if not validator.is_type(instance, "object"):
        return

    if not order:
        # propertyOrder may be an empty list
        return

    instance.property_order = order


def validate_flowStyle(validator, flow_style, instance, schema):
    """
    Sets a flag on the `tagged.TaggedList` or `tagged.TaggedDict`
    object so that the YAML generator knows which style to use to
    write the element.  In that sense this isn't really a "validator",
    but using the `jsonschema` library's extensible validation system
    is the easiest way to get this property assigned.
    """
    if not (validator.is_type(instance, "object") or validator.is_type(instance, "array")):
        return

    instance.flow_style = flow_style


def validate_style(validator, style, instance, schema):
    """
    Sets a flag on the `tagged.TaggedString` object so that the YAML
    generator knows which style to use to write the string.  In that
    sense this isn't really a "validator", but using the `jsonschema`
    library's extensible validation system is the easiest way to get
    this property assigned.
    """
    if not validator.is_type(instance, "string"):
        return

    instance.style = style


def validate_type(validator, types, instance, schema):
    """
    PyYAML returns strings that look like dates as datetime objects.
    However, as far as JSON is concerned, this is type==string and
    format==date-time.  That detects for that case and doesn't raise
    an error, otherwise falling back to the default type checker.
    """
    if isinstance(instance, datetime.datetime) and schema.get("format") == "date-time" and "string" in types:
        return None

    return mvalidators.Draft4Validator.VALIDATORS["type"](validator, types, instance, schema)


def validate_enum(validator, enums, instance, schema):
    """
    `asdf.tagged.Tagged` objects will fail in the default enum validator
    """

    if isinstance(instance, tagged.Tagged):
        instance = instance.base

    yield from mvalidators.Draft4Validator.VALIDATORS["enum"](validator, enums, instance, schema)


YAML_VALIDATORS = util.HashableDict(mvalidators.Draft4Validator.VALIDATORS.copy())
YAML_VALIDATORS.update(
    {
        "tag": validate_tag,
        "propertyOrder": validate_propertyOrder,
        "flowStyle": validate_flowStyle,
        "style": validate_style,
        "type": validate_type,
        "enum": validate_enum,
    },
)


def validate_fill_default(validator, properties, instance, schema):
    if not validator.is_type(instance, "object"):
        return

    for property_, subschema in properties.items():
        if "default" in subschema:
            instance.setdefault(property_, subschema["default"])

    yield from mvalidators.Draft4Validator.VALIDATORS["properties"](validator, properties, instance, schema)


FILL_DEFAULTS = util.HashableDict()
for key in ("allOf", "items"):
    FILL_DEFAULTS[key] = mvalidators.Draft4Validator.VALIDATORS[key]
FILL_DEFAULTS["properties"] = validate_fill_default


def validate_remove_default(validator, properties, instance, schema):
    if not validator.is_type(instance, "object"):
        return

    for property_, subschema in properties.items():
        if subschema.get("default", None) is not None and instance.get(property_, None) == subschema["default"]:
            del instance[property_]

    yield from mvalidators.Draft4Validator.VALIDATORS["properties"](validator, properties, instance, schema)


REMOVE_DEFAULTS = util.HashableDict()
for key in ("allOf", "items"):
    REMOVE_DEFAULTS[key] = mvalidators.Draft4Validator.VALIDATORS[key]
REMOVE_DEFAULTS["properties"] = validate_remove_default


class _ValidationContext:
    """
    Context that tracks (tree node, schema fragment) pairs that have
    already been validated.

    Instances of this class are context managers that track
    how many times they have been entered, and only reset themselves
    when exiting the outermost context.
    """

    __slots__ = ["_depth", "_seen"]

    def __init__(self):
        self._depth = 0
        self._seen = set()

    def add(self, instance, schema):
        """
        Inform the context that an instance has
        been validated against a schema fragment.
        """
        self._seen.add(self._make_seen_key(instance, schema))

    def seen(self, instance, schema):
        """
        Return True if an instance has already been
        validated against a schema fragment.
        """
        return self._make_seen_key(instance, schema) in self._seen

    def __enter__(self):
        self._depth += 1
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._depth -= 1

        if self._depth == 0:
            self._seen = set()

    def _make_seen_key(self, instance, schema):
        return (id(instance), id(schema))


@lru_cache
def _create_validator(validators=YAML_VALIDATORS, visit_repeat_nodes=False):
    meta_schema = _load_schema_cached(YAML_SCHEMA_METASCHEMA_ID, False)

    type_checker = mvalidators.Draft4Validator.TYPE_CHECKER.redefine_many(
        {
            "array": lambda checker, instance: isinstance(instance, (list, tuple)),
            "integer": lambda checker, instance: not isinstance(instance, bool) and isinstance(instance, Integral),
            "string": lambda checker, instance: isinstance(instance, (str, np.str_)),
        },
    )
    id_of = mvalidators.Draft4Validator.ID_OF
    ASDFvalidator = mvalidators.create(
        meta_schema=meta_schema,
        validators=validators,
        type_checker=type_checker,
        id_of=id_of,
    )

    def _patch_init(cls):
        original_init = cls.__init__

        def init(self, *args, **kwargs):
            self.ctx = kwargs.pop("ctx", None)
            self.serialization_context = kwargs.pop("serialization_context", None)

            original_init(self, *args, **kwargs)

        cls.__init__ = init

    def _patch_iter_errors(cls):
        original_iter_errors = cls.iter_errors

        cls._context = _ValidationContext()

        def iter_errors(self, instance, *args, **kwargs):
            # We can't validate anything that looks like an external reference,
            # since we don't have the actual content, so we just have to defer
            # it for now.  If the user cares about complete validation, they
            # can call `AsdfFile.resolve_references`.
            with self._context:
                if self._context.seen(instance, self.schema):
                    # We've already validated this instance against this schema,
                    # no need to do it again.
                    return

                if not visit_repeat_nodes:
                    self._context.add(instance, self.schema)

                if (isinstance(instance, dict) and "$ref" in instance) or isinstance(instance, reference.Reference):
                    return

                if not self.schema:
                    tag = tagged.get_tag(instance)
                    if tag is not None and self.serialization_context.extension_manager.handles_tag_definition(tag):
                        tag_def = self.serialization_context.extension_manager.get_tag_definition(tag)
                        schema_uris = tag_def.schema_uris

                        # Must validate against all schema_uris
                        for schema_uri in schema_uris:
                            try:
                                with self.resolver.resolving(schema_uri) as resolved:
                                    yield from self.descend(instance, resolved)
                            except RefResolutionError:
                                warnings.warn(f"Unable to locate schema file for '{tag}': '{schema_uri}'", AsdfWarning)

                    if isinstance(instance, dict):
                        for val in instance.values():
                            yield from self.iter_errors(val)

                    elif isinstance(instance, list):
                        for val in instance:
                            yield from self.iter_errors(val)
                else:
                    yield from original_iter_errors(self, instance)

        cls.iter_errors = iter_errors

    _patch_init(ASDFvalidator)
    _patch_iter_errors(ASDFvalidator)

    return ASDFvalidator


@lru_cache
def _load_schema(url):
    # handle case where the provided url is a non-str (Path)
    url = str(url)

    if url.startswith("http://") or url.startswith("https://") or url.startswith("asdf://"):
        msg = f"Unable to fetch schema from non-file URL: {url}"
        raise FileNotFoundError(msg)

    with generic_io.get_file(url) as fd:
        if isinstance(url, str) and url.endswith("json"):
            json_data = fd.read().decode("utf-8")
            result = json.loads(json_data, object_pairs_hook=OrderedDict)
        else:
            # The following call to yaml.load is safe because we're
            # using a loader that inherits from pyyaml's SafeLoader.
            result = yaml.load(fd, Loader=yamlutil.AsdfLoader)  # noqa: S506
    return result, fd.uri


def _make_schema_loader():
    def load_schema(url):
        # Check if this is a URI provided by the new
        # Mapping API:
        resource_manager = get_config().resource_manager

        if url in resource_manager:
            content = resource_manager[url]
            # The jsonschema metaschemas are JSON, but pyyaml
            # doesn't mind.
            # The following call to yaml.load is safe because we're
            # using a loader that inherits from pyyaml's SafeLoader.
            result = yaml.load(content, Loader=yamlutil.AsdfLoader)  # noqa: S506
            return result, url

        # If not, this must be a URL (or missing).  Fall back to fetching
        # the schema the old way:
        return _load_schema(url)

    return load_schema


def _make_jsonschema_refresolver():
    handlers = {}
    schema_loader = _make_schema_loader()

    def get_schema(url):
        return schema_loader(url)[0]

    for x in ["http", "https", "file", "tag", "asdf"]:
        handlers[x] = get_schema

    # Supplying our own implementation of urljoin_cache
    # allows asdf:// URIs to be resolved correctly.
    urljoin_cache = lru_cache(1024)(_patched_urllib_parse.urljoin)

    # We set cache_remote=False here because we do the caching of
    # remote schemas here in `load_schema`, so we don't need
    # jsonschema to do it on our behalf.  Setting it to `True`
    # counterintuitively makes things slower.
    return mvalidators.RefResolver(
        "",
        {},
        cache_remote=False,
        handlers=handlers,
        urljoin_cache=urljoin_cache,
    )


def load_schema(url, resolve_references=False):
    """
    Load a schema from the given URL.

    Parameters
    ----------
    url : str
        The path to the schema

    resolve_references : bool, optional
        If ``True``, resolve all ``$ref`` references.

    """
    # We want to cache the work that went into constructing the schema, but returning
    # the same object is treacherous, because users who mutate the result will not
    # expect that they're changing the schema everywhere.
    return copy.deepcopy(_load_schema_cached(url, resolve_references))


def _safe_resolve(json_id, uri):
    """
    This function handles the tricky task of resolving a schema URI
    in the presence of both new and legacy extensions.

    There are two senses of "resolve" here: one is to resolve the URI
    to a file:// URL using the legacy extension resolver object. The other
    is to resolve relative URIs against the id of the current schema document,
    which is what generic_io.resolve_uri does.

    For URIs associated with new-style extensions, we want to resolve with
    generic_io.resolve_uri, but not with the resolver object, otherwise we risk
    mangling URIs that share a prefix with a resolver mapping.
    """
    # We can't use urllib.parse here because tag: URIs don't
    # parse correctly.
    parts = uri.split("#")
    base = parts[0]
    fragment = parts[1] if len(parts) > 1 else ""

    # Resolve relative URIs (e.g., #foo/bar, ../foo/bar) against
    # the current schema id.
    base = generic_io.resolve_uri(json_id, base)

    return base, fragment


@lru_cache
def _load_schema_cached(url, resolve_references):
    loader = _make_schema_loader()
    schema, url = loader(url)

    if resolve_references:

        def resolve_refs(node, json_id):
            if json_id is None:
                json_id = url

            if isinstance(node, dict) and "$ref" in node:
                suburl_base, suburl_fragment = _safe_resolve(json_id, node["$ref"])

                if suburl_base == url or suburl_base == schema.get("id"):
                    # This is a local ref, which we'll resolve in both cases.
                    subschema = schema
                else:
                    subschema = load_schema(suburl_base, True)

                return reference.resolve_fragment(subschema, suburl_fragment)

            return node

        schema = treeutil.walk_and_modify(schema, resolve_refs)

    return schema


def get_validator(
    schema=None,
    ctx=None,
    validators=None,
    *args,
    _visit_repeat_nodes=False,
    _serialization_context=None,
    **kwargs,
):
    """
    Get a JSON schema validator object for the given schema.

    The additional *args and **kwargs are passed to the
    constructor of the returned ``Validator``.

    Parameters
    ----------
    schema : schema, optional
        Explicit schema to use.  If not provided, the schema to use
        is determined by the tag on instance (or subinstance).

    ctx : AsdfFile context
        Used to resolve tags and urls

    validators : dict, optional
        A dictionary mapping properties to validators to use (instead
        of the built-in ones and ones provided by extension types).

    _visit_repeat_nodes : bool, optional
        Force the validator to visit nodes that it has already
        seen.  This flag is a temporary hack to support a specific
        project that uses a custom validator to update a .fits file.
        Setting `True` is discouraged and will lead to RecursionError
        in trees containing reference cycles.

    Returns
    -------
    validator : jsonschema.Validator
    """
    if ctx is None:
        from ._asdf import AsdfFile

        ctx = AsdfFile()

    if _serialization_context is None:
        _serialization_context = ctx._create_serialization_context()

    if validators is None:
        validators = util.HashableDict(YAML_VALIDATORS.copy())
        validators.update(ctx._extension_manager.validator_manager.get_jsonschema_validators())

    kwargs["resolver"] = _make_jsonschema_refresolver()

    # We don't just call validators.validate() directly here, because
    # that validates the schema itself, wasting a lot of time (at the
    # time of this writing, it was half of the runtime of the unit
    # test suite!!!).  Instead, we assume that the schemas are valid
    # through the running of the unit tests, not at run time.
    cls = _create_validator(validators=validators, visit_repeat_nodes=_visit_repeat_nodes)
    return cls({} if schema is None else schema, *args, ctx=ctx, serialization_context=_serialization_context, **kwargs)


def _validate_large_literals(instance, reading):
    """
    Validate that the tree has no large numeric literals.
    """

    def _validate(value):
        if value <= constants.MAX_NUMBER and value >= constants.MIN_NUMBER:
            return

        if reading:
            warnings.warn(
                f"Invalid integer literal value {value} detected while reading file. "
                "The value has been read safely, but the file should be "
                "fixed.",
                AsdfWarning,
            )
        else:
            msg = f"Integer value {value} is too large to safely represent as a literal in ASDF"
            raise ValidationError(msg)

    if isinstance(instance, Integral):
        _validate(instance)
    elif isinstance(instance, Mapping):
        for key in instance:
            if isinstance(key, Integral):
                _validate(key)


def _validate_mapping_keys(instance, reading):
    """
    Validate that mappings do not contain illegal key types
    (as of ASDF core schemas 1.6.0, only str, int, and bool are
    permitted).
    """
    if not isinstance(instance, Mapping):
        return

    for key in instance:
        if isinstance(key, tagged.Tagged) or not isinstance(key, (str, int, bool)):
            if reading:
                warnings.warn(
                    f"Invalid mapping key {key} detected while reading file. "
                    "The value has been read safely, but the file should be "
                    "fixed.",
                    AsdfWarning,
                )
            else:
                msg = f"Mapping key {key} is not permitted.  Valid types: str, int, bool."
                raise ValidationError(msg)


def validate(instance, ctx=None, schema=None, validators=None, reading=False, *args, **kwargs):
    """
    Validate the given instance (which must be a tagged tree) against
    the appropriate schema.  The schema itself is located using the
    tag on the instance.

    The additional ``*args`` and ``**kwargs`` are passed along to
    `~jsonschema.protocols.Validator.validate`.

    Parameters
    ----------
    instance : tagged tree

    ctx : AsdfFile context
        Used to resolve tags and urls

    schema : schema, optional
        Explicit schema to use.  If not provided, the schema to use
        is determined by the tag on instance (or subinstance) and
        any custom schema provided to ``ctx``.

    validators : dict, optional
        A dictionary mapping properties to validators to use (instead
        of the built-in ones and ones provided by extension types).

    reading: bool, optional
        Indicates whether validation is being performed when the file is being
        read. This is useful to allow for different validation behavior when
        reading vs writing files.
    """
    if ctx is None:
        from ._asdf import AsdfFile

        ctx = AsdfFile()

    if schema is None and ctx._custom_schema:
        schema = ctx._custom_schema
    validator = get_validator({} if schema is None else schema, ctx, validators, *args, **kwargs)
    validator.validate(instance)

    additional_validators = [_validate_large_literals]
    if ctx.version >= versioning.RESTRICTED_KEYS_MIN_VERSION:
        additional_validators.append(_validate_mapping_keys)

    def _callback(instance):
        for validator in additional_validators:
            validator(instance, reading)

    treeutil.walk(instance, _callback)


def fill_defaults(instance, ctx, reading=False):
    """
    For any default values in the schema, add them to the tree if they
    don't exist.

    Parameters
    ----------
    instance : tagged tree

    ctx : AsdfFile context
        Used to resolve tags and urls

    reading: bool, optional
        Indicates whether the ASDF file is being read (in contrast to being
        written).
    """
    validate(instance, ctx, validators=FILL_DEFAULTS, reading=reading)


def remove_defaults(instance, ctx):
    """
    For any values in the tree that are the same as the default values
    specified in the schema, remove them from the tree.

    Parameters
    ----------
    instance : tagged tree

    ctx : AsdfFile context
        Used to resolve tags and urls
    """
    validate(instance, ctx, validators=REMOVE_DEFAULTS)


def check_schema(schema, validate_default=True):
    """
    Check a given schema to make sure it is valid YAML schema.

    Parameters
    ----------
    schema : dict
        The schema object, as returned by ``load_schema``.
    validate_default : bool, optional
        Set to `True` to validate the content of the default
        field against the schema.
    """
    validators = util.HashableDict(mvalidators.Draft4Validator.VALIDATORS.copy())

    if validate_default:
        # The jsonschema library doesn't validate defaults
        # on its own.
        instance_validator = get_validator(schema)
        instance_scope = schema.get("id", "")

        def _validate_default(validator, default, instance, schema):
            if not validator.is_type(instance, "object"):
                return

            if "default" in instance:
                instance_validator.resolver.push_scope(instance_scope)
                try:
                    yield from instance_validator.descend(instance["default"], instance)
                finally:
                    instance_validator.resolver.pop_scope()

        validators.update({"default": _validate_default})

        def applicable_validators(schema):
            items = list(schema.items())
            items.append(("default", ""))
            return items

    else:
        applicable_validators = methodcaller("items")

    meta_schema_id = schema.get("$schema", YAML_SCHEMA_METASCHEMA_ID)
    meta_schema = _load_schema_cached(meta_schema_id, False)

    resolver = _make_jsonschema_refresolver()

    cls = mvalidators.create(
        meta_schema=meta_schema,
        validators=validators,
        type_checker=mvalidators.Draft4Validator.TYPE_CHECKER,
        id_of=mvalidators.Draft4Validator.ID_OF,
        applicable_validators=applicable_validators,
    )
    validator = cls(meta_schema, resolver=resolver)
    validator.validate(schema)



import copy
import datetime
import io
import os
import time
import warnings
import weakref

from packaging.version import Version

from . import _compression as mcompression
from . import _display as display
from . import _io, constants, generic_io, lazy_nodes, reference, schema, treeutil, util, versioning, yamlutil
from . import _node_info as node_info
from ._block.manager import Manager as BlockManager
from ._helpers import validate_version
from .config import config_context, get_config
from .exceptions import (
    AsdfManifestURIMismatchWarning,
    AsdfPackageVersionWarning,
    AsdfWarning,
)
from .extension import Extension, ExtensionProxy, _serialization_context, get_cached_extension_manager
from .search import AsdfSearchResult
from .tags.core import AsdfObject, ExtensionMetadata, HistoryEntry, Software
from .util import NotSet


class AsdfFile:
    """
    The main class that represents an ASDF file object.
    """

    def __init__(
        self,
        tree=None,
        uri=None,
        extensions=None,
        version=None,
        ignore_unrecognized_tag=False,
        memmap=False,
        lazy_load=True,
        custom_schema=None,
    ):
        """
        Parameters
        ----------
        tree : dict or AsdfFile, optional
            The main tree data in the ASDF file.  Must conform to the
            ASDF schema.

        uri : str, optional
            The URI for this ASDF file.  Used to resolve relative
            references against.  If not provided, will be
            automatically determined from the associated file object,
            if possible and if created from `asdf.open`.

        extensions : object, optional
            Additional extensions to use when reading and writing the file.
            May be an `asdf.extension.Extension` or a `list` of extensions.

        version : str, optional
            The ASDF core schemas version.  If not provided, defaults to the
            configured default version.  See `asdf.config.AsdfConfig.default_version`.

        ignore_unrecognized_tag : bool, optional
            When `True`, do not raise warnings for unrecognized tags. Set to
            `False` by default.

        memmap : bool, optional
            When `True`, when reading files, attempt to memmap underlying data
            arrays when possible. Defaults to ``False``.

        lazy_load : bool, optional
            When `True` and the underlying file handle is seekable, data
            arrays will only be loaded lazily: i.e. when they are accessed
            for the first time. In this case the underlying file must stay
            open during the lifetime of the tree. Setting to False causes
            all data arrays to be loaded up front, which means that they
            can be accessed even after the underlying file is closed.

        custom_schema : str, optional
            Path to a custom schema file that will be used for a secondary
            validation pass. This can be used to ensure that particular ASDF
            files follow custom conventions beyond those enforced by the
            specification.
        """
        # make a new AsdfVersion instance here so files don't share the same instance
        self._file_format_version = versioning.AsdfVersion(versioning._FILE_FORMAT_VERSION)

        # Don't use the version setter here; it tries to access
        # the extensions, which haven't been assigned yet.
        if version is None:
            self._version = versioning.AsdfVersion(get_config().default_version)
        else:
            self._version = versioning.AsdfVersion(validate_version(version))

        self._user_extensions = self._process_user_extensions(extensions)
        self._plugin_extensions = self._process_plugin_extensions()
        self._extension_manager = None

        if custom_schema is not None:
            self._custom_schema = schema._load_schema_cached(custom_schema, True)
        else:
            self._custom_schema = None

        self._ignore_unrecognized_tag = ignore_unrecognized_tag

        # Context of a call to treeutil.walk_and_modify, needed in the AsdfFile
        # in case walk_and_modify is re-entered by extension code (via
        # custom_tree_to_tagged_tree or tagged_tree_to_custom_tree).
        self._tree_modification_context = treeutil._TreeModificationContext()

        # A cache of tagged objects and their converted custom objects used when
        # a file is read with "lazy_tree=True". Used by lazy_nodes.
        self._tagged_object_cache = lazy_nodes._TaggedObjectCache()

        self._fd = None
        self._closed = False
        self._external_asdf_by_uri = {}
        self._blocks = BlockManager(uri=uri, lazy_load=lazy_load, memmap=memmap)
        if tree is None:
            # Bypassing the tree property here, to avoid validating
            # an empty tree.
            self._tree = AsdfObject()
        elif isinstance(tree, AsdfFile):
            if self.extensions != tree.extensions:
                # TODO(eslavich): Why not?  What if that's the goal
                # of copying the file?
                msg = "Can not copy AsdfFile and change active extensions"
                raise ValueError(msg)
            self._blocks._uri = tree.uri
            # Set directly to self._tree (bypassing property), since
            # we can assume the other AsdfFile is already valid.
            # Call "walk_and_modify" here to use it's logic for
            # creating a "copy" of the other tree. This mimics what was previously
            # a call to find_references (which we longer do in AsdfFile.__init__)
            self._tree = treeutil.walk_and_modify(tree.tree, lambda o: o)
        else:
            self._tree = AsdfObject(tree)

        self._comments = []

    @property
    def version(self):
        """
        Get this AsdfFile's ASDF core schemas version.

        Returns
        -------
        asdf.versioning.AsdfVersion
        """
        return self._version

    @version.setter
    def version(self, value):
        """
        Set this AsdfFile's ASDF core schemas version.

        Parameters
        ----------
        value : str or asdf.versioning.AsdfVersion
        """
        self._version = versioning.AsdfVersion(validate_version(value))
        # The new version may not be compatible with the previous
        # set of extensions, so we need to check them again:
        self._user_extensions = self._process_user_extensions(self._user_extensions)
        self._plugin_extensions = self._process_plugin_extensions()
        self._extension_manager = None

    @property
    def version_string(self):
        """
        Get this AsdfFile's ASDF core schemas version as a string.

        Returns
        -------
        str
        """
        return str(self._version)

    @property
    def extensions(self):
        """
        Get the list of user extensions that are enabled for
        use with this AsdfFile.

        Returns
        -------
        list of asdf.extension.ExtensionProxy
        """
        return self._user_extensions

    @extensions.setter
    def extensions(self, value):
        """
        Set the list of user extensions that are enabled for
        use with this AsdfFile.

        Parameters
        ----------
        value : list of asdf.extension.Extension
        """
        self._user_extensions = self._process_user_extensions(value)
        self._extension_manager = None

    @property
    def extension_manager(self):
        """
        Get the ExtensionManager for this AsdfFile.

        Returns
        -------
        asdf.extension.ExtensionManager
        """
        if self._extension_manager is None:
            self._extension_manager = get_cached_extension_manager(self._user_extensions + self._plugin_extensions)
        return self._extension_manager

    def __enter__(self):
        return self

    def __exit__(self, type_, value, traceback):
        self.close()

    def _check_extensions(self, tree, strict=False):
        """
        Compare the user's installed extensions to metadata in the tree
        and warn when a) an extension is missing or b) an extension is
        present but the file was written with a later version of the
        extension's package.

        Parameters
        ----------
        tree : AsdfObject
            Fully converted tree of custom types.
        strict : bool, optional
            Set to `True` to convert warnings to exceptions.
        """
        if (
            "history" not in tree
            or not isinstance(tree["history"], (dict, lazy_nodes.AsdfDictNode))
            or "extensions" not in tree["history"]
        ):
            return

        for extension in tree["history"]["extensions"]:
            installed = None
            for ext in self._user_extensions + self._plugin_extensions:
                if (extension.extension_uri is not None and extension.extension_uri == ext.extension_uri) or (
                    extension.extension_uri is None and extension.extension_class in ext.legacy_class_names
                ):
                    installed = ext
                    break

            filename = f"'{self._fd._uri}'" if (self._fd is not None and self._fd._uri) else ""
            if extension.extension_uri is not None:
                extension_description = f"URI '{extension.extension_uri}'"
            else:
                extension_description = f"class '{extension.extension_class}'"
            if extension.software is not None:
                extension_description += (
                    f" (from package {extension.software['name']}=={extension.software['version']})"
                )

            if installed is None:
                msg = (
                    f"File {filename}was created with extension "
                    f"{extension_description}, which is not currently installed"
                )
                if strict:
                    raise RuntimeError(msg)

                warnings.warn(msg, AsdfPackageVersionWarning)

            elif extension.software:
                # Local extensions may not have a real version.  If the package name changed,
                # then the version sequence may have been reset.
                if installed.package_version is None or installed.package_name != extension.software["name"]:
                    continue
                # Compare version in file metadata with installed version
                if Version(installed.package_version) < Version(extension.software["version"]):
                    msg = (
                        f"File {filename}was created with extension {extension_description}, "
                        f"but older package ({installed.package_name}=={installed.package_version}) is installed."
                    )
                    if strict:
                        raise RuntimeError(msg)

                    warnings.warn(msg, AsdfPackageVersionWarning)

            # check version of manifest providing package (if one was recorded)
            if "manifest_software" in extension:
                package_name = extension["manifest_software"]["name"]
                package_version = Version(extension["manifest_software"]["version"])
                package_description = f"{package_name}=={package_version}"
                installed_version = None
                for mapping in get_config().resource_manager._resource_mappings:
                    if mapping.package_name == package_name:
                        installed_version = Version(mapping.package_version)
                        break
                msg = None
                if installed_version is None:
                    msg = (
                        f"File {filename}was created with package {package_description}, "
                        "which is currently not installed"
                    )
                elif installed_version < package_version:
                    msg = (
                        f"File {filename}was created with package {package_description}, "
                        f"but older package({package_name}=={installed_version}) is installed."
                    )
                if msg:
                    if strict:
                        raise RuntimeError(msg)
                    warnings.warn(msg, AsdfPackageVersionWarning)

    def _process_plugin_extensions(self):
        """
        Select installed extensions that are compatible with this
        file's ASDF core schemas version.

        Returns
        -------
        list of asdf.extension.ExtensionProxy
        """
        return [e for e in get_config().extensions if self.version_string in e.asdf_standard_requirement]

    def _process_user_extensions(self, extensions):
        """
        Validate a list of extensions requested by the user
        add missing extensions registered with the current `AsdfConfig`.

        Parameters
        ----------
        extensions : object
            May be an `asdf.extension.Extension` or a `list` of extensions.

        Returns
        -------
        list of asdf.extension.ExtensionProxy
        """
        if extensions is None:
            extensions = []
        elif isinstance(extensions, (Extension, ExtensionProxy)):
            extensions = [extensions]

        if not isinstance(extensions, list):
            msg = "The extensions parameter must be an extension or list of extensions"
            raise TypeError(msg)

        extensions = [ExtensionProxy.maybe_wrap(e) for e in extensions]

        result = []
        for extension in extensions:
            if self.version_string not in extension.asdf_standard_requirement:
                warnings.warn(
                    f"Extension {extension} does not support ASDF Standard {self.version_string}.  "
                    "It has been disabled.",
                    AsdfWarning,
                )
            else:
                result.append(extension)

        return result

    def _update_extension_history(self, tree, serialization_context):
        """
        Update the extension metadata on this file's tree to reflect
        extensions used during serialization.

        Parameters
        ----------
        serialization_context : asdf.extension.SerializationContext
            The context that was used to serialize the tree.
        """
        if serialization_context.version < versioning.NEW_HISTORY_FORMAT_MIN_VERSION:
            return

        if "history" not in tree:
            tree["history"] = {"extensions": []}
        # Support clients who are still using the old history format
        elif isinstance(tree["history"], list):
            histlist = tree["history"]
            tree["history"] = {"entries": histlist, "extensions": []}
            warnings.warn(
                "The ASDF history format has changed in order to "
                "support metadata about extensions. History entries "
                "should now be stored under tree['history']['entries'].",
                AsdfWarning,
            )
        elif "extensions" not in tree["history"]:
            tree["history"]["extensions"] = []

        for extension in serialization_context._extensions_used:
            ext_name = extension.class_name
            ext_meta = ExtensionMetadata(extension_class=ext_name)
            if extension.package_name is not None:
                ext_meta["software"] = Software(name=extension.package_name, version=extension.package_version)
            if extension.extension_uri is not None:
                ext_meta["extension_uri"] = extension.extension_uri
            if extension.compressors:
                ext_meta["supported_compression"] = [comp.label.decode("ascii") for comp in extension.compressors]
            manifest = getattr(extension._delegate, "_manifest", None)
            if manifest is not None:
                # check if this extension was built from a manifest is a different package
                resource_mapping = get_config().resource_manager._mappings_by_uri.get(manifest["id"])
                # if an extension was registered with a manifest uri that does not match the id the
                # resource_mapping lookup will fail (resource_mapping will be None)
                if resource_mapping is None:
                    warnings.warn(
                        f"Extension ({extension.extension_uri}) uses a manifest with a uri that "
                        "does not match it's id. Please open an issue with the extension maintainer",
                        AsdfManifestURIMismatchWarning,
                    )
                elif resource_mapping.package_name != extension.package_name:
                    ext_meta["manifest_software"] = Software(
                        name=resource_mapping.package_name,
                        version=resource_mapping.package_version,
                    )

            for i, entry in enumerate(tree["history"]["extensions"]):
                # Update metadata about this extension if it already exists
                if (
                    entry.extension_uri is not None and entry.extension_uri == extension.extension_uri
                ) or entry.extension_class in extension.legacy_class_names:
                    tree["history"]["extensions"][i] = ext_meta
                    break
            else:
                tree["history"]["extensions"].append(ext_meta)

    @property
    def file_format_version(self):
        return self._file_format_version

    def close(self):
        """
        Close the file handles associated with the `asdf.AsdfFile`.
        """
        if self._fd and not self._closed:
            # This is ok to always do because GenericFile knows
            # whether it "owns" the file and should close it.
            self._fd.close()
            self._fd = None
            self._closed = True
            # as we're closing the file, also empty out the
            # tree so that references to array data can be released
            self._tree = AsdfObject()
            self._tagged_object_cache.clear()
        for external in self._external_asdf_by_uri.values():
            external.close()
        self._external_asdf_by_uri.clear()
        self._blocks.close()

    def copy(self):
        return self.__class__(
            copy.deepcopy(self._tree),
            self._blocks._uri,
            self._user_extensions,
        )

    __copy__ = __deepcopy__ = copy

    @property
    def uri(self):
        """
        Get the URI associated with the `AsdfFile`.

        In many cases, it is automatically determined from the file
        handle used to read or write the file.
        """
        return self._blocks._uri

    def resolve_uri(self, uri):
        """
        Resolve a (possibly relative) URI against the URI of this ASDF
        file.  May be overridden by base classes to change how URIs
        are resolved.  This does not apply any ``uri_mapping`` that was
        passed to the constructor.

        Parameters
        ----------
        uri : str
            An absolute or relative URI to resolve against the URI of
            this ASDF file.

        Returns
        -------
        uri : str
            The resolved URI.
        """
        return generic_io.resolve_uri(self.uri, uri)

    def open_external(self, uri, **kwargs):
        """
        Open an external ASDF file, from the given (possibly relative)
        URI.  There is a cache (internal to this ASDF file) that ensures
        each external ASDF file is loaded only once.

        Parameters
        ----------
        uri : str
            An absolute or relative URI to resolve against the URI of
            this ASDF file.

        Returns
        -------
        asdffile : AsdfFile
            The external ASDF file.
        """
        # For a cache key, we want to ignore the "fragment" part.
        base_uri = util.get_base_uri(uri)
        resolved_uri = self.resolve_uri(base_uri)

        # A uri like "#" should resolve back to ourself.  In that case,
        # just return `self`.
        if resolved_uri == "" or resolved_uri == self.uri:
            return self

        asdffile = self._external_asdf_by_uri.get(resolved_uri)
        if asdffile is None:
            asdffile = open_asdf(resolved_uri, mode="r", **kwargs)
            self._external_asdf_by_uri[resolved_uri] = asdffile
        return asdffile

    @property
    def tree(self):
        """
        Get/set the tree of data in the ASDF file.

        When set, the tree will be validated against the ASDF schema.
        """
        if self._closed:
            msg = "Cannot access data from closed ASDF file"
            raise OSError(msg)
        return self._tree

    @tree.setter
    def tree(self, tree):
        self._tree = AsdfObject(tree)

    def keys(self):
        return self.tree.keys()

    def __getitem__(self, key):
        return self.tree[key]

    def __setitem__(self, key, value):
        self.tree[key] = value

    def __contains__(self, item):
        return item in self.tree

    @property
    def comments(self):
        """
        Get the comments after the header, before the tree.
        """
        return self._comments

    def _validate(self, tree, reading=False):
        with self._blocks.options_context():
            # If we're validating on read then the tree
            # is already guaranteed to be in tagged form.
            tagged_tree = tree if reading else yamlutil.custom_tree_to_tagged_tree(tree, self)

            schema.validate(tagged_tree, self, reading=reading)

    def validate(self):
        """
        Validate the current state of the tree against the ASDF schema.
        """
        self._validate(self._tree)

    def make_reference(self, path=None):
        """
        Make a new reference to a part of this file's tree, that can be
        assigned as a reference to another tree.

        Parameters
        ----------
        path : list of str and int, optional
            The parts of the path pointing to an item in this tree.
            If omitted, points to the root of the tree.

        Returns
        -------
        reference :
            A reference object.

        Examples
        --------
        For the given AsdfFile ``ff``, add an external reference to the data in
        an external file::

            >>> import asdf
            >>> flat = asdf.open("http://stsci.edu/reference_files/flat.asdf")  # doctest: +SKIP
            >>> ff.tree['flat_field'] = flat.make_reference(['data'])  # doctest: +SKIP
        """
        return reference.make_reference(self, [] if path is None else path)

    def set_array_storage(self, arr, array_storage):
        """
        Set the block type to use for the given array data.

        Parameters
        ----------
        arr : numpy.ndarray
            The array to set.  If multiple views of the array are in
            the tree, only the most recent block type setting will be
            used, since all views share a single block.

        array_storage : str
            Must be one of:

            - ``internal``: The default.  The array data will be
              stored in a binary block in the same ASDF file.

            - ``external``: Store the data in a binary block in a
              separate ASDF file.

            - ``inline``: Store the data as YAML inline in the tree.
        """
        self._blocks._set_array_storage(arr, array_storage)

    def get_array_storage(self, arr):
        """
        Get the block type for the given array data.

        Parameters
        ----------
        arr : numpy.ndarray
        """
        return self._blocks._get_array_storage(arr)

    def set_array_compression(self, arr, compression, **compression_kwargs):
        """
        Set the compression to use for the given array data.

        Parameters
        ----------
        arr : numpy.ndarray
            The array to set.  If multiple views of the array are in
            the tree, only the most recent compression setting will be
            used, since all views share a single block.

        compression : str or None
            Must be one of:

            - ``''`` or `None`: no compression

            - ``zlib``: Use zlib compression

            - ``bzp2``: Use bzip2 compression

            - ``lz4``: Use lz4 compression

            - ``input``: Use the same compression as in the file read.
              If there is no prior file, acts as None.

        """
        self._blocks._set_array_compression(arr, compression, **compression_kwargs)

    def get_array_compression(self, arr):
        """
        Get the compression type for the given array data.

        Parameters
        ----------
        arr : numpy.ndarray

        Returns
        -------
        compression : str or None
        """
        return self._blocks._get_array_compression(arr)

    def get_array_compression_kwargs(self, arr):
        """ """
        return self._blocks._get_array_compression_kwargs(arr)

    def set_array_save_base(self, arr, save_base):
        """
        Set the ``save_base`` option for ``arr``. When ``arr`` is
        written to a file, if ``save_base`` is ``True`` the base array
        for ``arr`` will be saved.

        Note that similar to other array options this setting is linked
        to the base array if ``arr`` is a view.

        Parameters
        ----------
        arr : numpy.ndarray

        save_base : bool or None
            if ``None`` the ``default_array_save_base`` value from asdf
            config will be used
        """
        self._blocks._set_array_save_base(arr, save_base)

    def get_array_save_base(self, arr):
        """
        Returns the ``save_base`` option for ``arr``. When ``arr`` is
        written to a file, if ``save_base`` is ``True`` the base array
        for ``arr`` will be saved.

        Parameters
        ----------
        arr : numpy.ndarray

        Returns
        -------
        save_base : bool or None
        """
        return self._blocks._get_array_save_base(arr)

    def _write_tree(self, tree, fd, pad_blocks):
        fd.write(constants.ASDF_MAGIC)
        fd.write(b" ")
        fd.write(f"{self.file_format_version}".encode("ascii"))
        fd.write(b"\n")

        fd.write(b"#")
        fd.write(constants.ASDF_STANDARD_COMMENT)
        fd.write(b" ")
        fd.write(self.version_string.encode("ascii"))
        fd.write(b"\n")

        if len(tree):
            serialization_context = self._create_serialization_context(_serialization_context.BlockAccess.WRITE)

            for compression in self._blocks.get_output_compressions():
                # lookup extension
                compressor = mcompression._get_compressor_from_extensions(compression, return_extension=True)
                if compressor is not None:
                    # mark it as used
                    serialization_context._mark_extension_used(compressor[1])

            def _tree_finalizer(tagged_tree):
                """
                The list of extensions used is not known until after
                serialization, so we're using a hook provided by
                yamlutil.dump_tree to update extension metadata
                after the tree has been converted to tagged objects.
                """
                self._update_extension_history(tree, serialization_context)
                if "history" in tree:
                    tagged_tree["history"] = yamlutil.custom_tree_to_tagged_tree(
                        tree["history"],
                        self,
                        _serialization_context=serialization_context,
                    )
                else:
                    tagged_tree.pop("history", None)

            yamlutil.dump_tree(
                tree,
                fd,
                self,
                tree_finalizer=_tree_finalizer,
                _serialization_context=serialization_context,
            )

        if pad_blocks:
            padding = util.calculate_padding(fd.tell(), pad_blocks, fd.block_size)
            fd.fast_forward(padding)

    def _serial_write(self, fd, pad_blocks, include_block_index):
        with self._blocks.write_context(fd):
            # prep a tree for a writing
            tree = copy.copy(self._tree)
            tree["asdf_library"] = _io.get_asdf_library_info()
            if "history" in self._tree:
                tree["history"] = copy.deepcopy(self._tree["history"])

            self._write_tree(tree, fd, pad_blocks)
            self._blocks.write(pad_blocks, include_block_index)

    def update(
        self,
        all_array_storage=NotSet,
        all_array_compression=NotSet,
        compression_kwargs=NotSet,
        pad_blocks=False,
        include_block_index=True,
        version=None,
    ):
        """
        Update the file on disk in place.

        Parameters
        ----------
        all_array_storage : string, optional
            If provided, override the array storage type of all blocks
            in the file immediately before writing.  Must be one of:

            - ``internal``: The default.  The array data will be
              stored in a binary block in the same ASDF file.

            - ``external``: Store the data in a binary block in a
              separate ASDF file.

            - ``inline``: Store the data as YAML inline in the tree.

        all_array_compression : string, optional
            If provided, set the compression type on all binary blocks
            in the file.  Must be one of:

            - ``''`` or `None`: No compression.

            - ``zlib``: Use zlib compression.

            - ``bzp2``: Use bzip2 compression.

            - ``lz4``: Use lz4 compression.

            - ``input``: Use the same compression as in the file read.
              If there is no prior file, acts as None

        compression_kwargs : dict, optional
            If provided, set this as the compression keyword arguments
            for all binary blocks in the file.

        pad_blocks : float or bool, optional
            Add extra space between blocks to allow for updating of
            the file.  If `False` (default), add no padding (always
            return 0).  If `True`, add a default amount of padding of
            10% If a float, it is a factor to multiple content_size by
            to get the new total size.

        include_block_index : bool, optional
            If `False`, don't include a block index at the end of the
            file.  (Default: `True`)  A block index is never written
            if the file has a streamed block.

        version : str, optional
            Update the ASDF core schemas version of this AsdfFile before
            writing.
        """

        with config_context() as config:
            if all_array_storage is not NotSet:
                config.all_array_storage = all_array_storage
            if all_array_compression is not NotSet:
                config.all_array_compression = all_array_compression
            if compression_kwargs is not NotSet:
                config.all_array_compression_kwargs = compression_kwargs

            fd = self._fd

            if fd is None:
                msg = "Can not update, since there is no associated file"
                raise ValueError(msg)

            if not fd.writable():
                msg = (
                    "Can not update, since associated file is read-only. Make "
                    "sure that the AsdfFile was opened with mode='rw' and the "
                    "underlying file handle is writable."
                )
                raise OSError(msg)

            if not fd.seekable():
                msg = "Can not update, since associated file is not seekable"
                raise OSError(msg)

            if version is not None:
                self.version = version

            # flush all pending memmap writes
            if fd.can_memmap():
                fd.flush_memmap()

            def rewrite():
                self._fd.seek(0)
                self._serial_write(self._fd, pad_blocks, include_block_index)
                self._fd.truncate()
                if self._fd.can_memmap():
                    self._fd.close_memmap()

            # if we have no read blocks, we can just call write_to as no internal blocks are reused
            if len(self._blocks.blocks) == 0:
                rewrite()
                return

            # if we have all external blocks, we can just call write_to as no internal blocks are reused
            if config.all_array_storage == "external":
                rewrite()
                return

            self._tree["asdf_library"] = _io.get_asdf_library_info()

            # prepare block manager for writing
            with self._blocks.write_context(self._fd, copy_options=False):
                # write out tree to temporary buffer
                tree_fd = generic_io.get_file(io.BytesIO(), mode="rw")
                self._write_tree(self._tree, tree_fd, False)
                new_tree_size = tree_fd.tell()

                # update blocks
                self._blocks.update(new_tree_size, pad_blocks, include_block_index)
                end_of_file = self._fd.tell()

            # now write the tree
            self._fd.seek(0)
            tree_fd.seek(0)
            self._fd.write(tree_fd.read())
            self._fd.flush()

            # close memmap to trigger arrays to reload themselves
            self._fd.seek(end_of_file)
            self._fd.truncate()
            if self._fd.can_memmap():
                self._fd.close_memmap()

    def write_to(
        self,
        fd,
        all_array_storage=NotSet,
        all_array_compression=NotSet,
        compression_kwargs=NotSet,
        pad_blocks=False,
        include_block_index=True,
        version=None,
    ):
        """
        Write the ASDF file to the given file-like object.

        `write_to` does not change the underlying file descriptor in
        the `asdf.AsdfFile` object, but merely copies the content to a new
        file.

        Parameters
        ----------
        fd : string or file-like object
            May be a string path to a file, or a Python file-like
            object.  If a string path, the file is automatically
            closed after writing.  If not a string path, it is the
            caller's responsibility to close the object.

        all_array_storage : string, optional
            If provided, override the array storage type of all blocks
            in the file immediately before writing.  Must be one of:

            - ``internal``: The default.  The array data will be
              stored in a binary block in the same ASDF file.

            - ``external``: Store the data in a binary block in a
              separate ASDF file.

            - ``inline``: Store the data as YAML inline in the tree.

        all_array_compression : string, optional
            If provided, set the compression type on all binary blocks
            in the file.  Must be one of:

            - ``''`` or `None`: No compression.

            - ``zlib``: Use zlib compression.

            - ``bzp2``: Use bzip2 compression.

            - ``lz4``: Use lz4 compression.

            - ``input``: Use the same compression as in the file read.
              If there is no prior file, acts as None.

        compression_kwargs : dict, optional
            If provided, set this as the compression keyword arguments
            for all binary blocks in the file.

        pad_blocks : float or bool, optional
            Add extra space between blocks to allow for updating of
            the file.  If `False` (default), add no padding (always
            return 0).  If `True`, add a default amount of padding of
            10% If a float, it is a factor to multiple content_size by
            to get the new total size.

        include_block_index : bool, optional
            If `False`, don't include a block index at the end of the
            file.  (Default: `True`)  A block index is never written
            if the file has a streamed block.

        version : str, optional
            Update the ASDF core schemas version of this AsdfFile before
            writing.
        """
        with config_context() as config:
            if all_array_storage is not NotSet:
                config.all_array_storage = all_array_storage
            if all_array_compression is not NotSet:
                config.all_array_compression = all_array_compression
            if compression_kwargs is not NotSet:
                config.all_array_compression_kwargs = compression_kwargs

            if version is not None:
                previous_version = self.version
                self.version = version

            try:
                with generic_io.get_file(fd, mode="w") as fd:
                    self._serial_write(fd, pad_blocks, include_block_index)
            finally:
                if version is not None:
                    self.version = previous_version

    def find_references(self):
        """
        Finds all external "JSON References" in the tree and converts
        them to ``reference.Reference`` objects.
        """
        # Set directly to self._tree, since it doesn't need to be re-validated.
        self._tree = reference.find_references(self._tree, self)

    def resolve_references(self):
        """
        Finds all external "JSON References" in the tree, loads the
        external content, and places it directly in the tree.  Saving
        a ASDF file after this operation means it will have no
        external references, and will be completely self-contained.
        """
        self._tree = reference.resolve_references(self._tree, self)

    def fill_defaults(self):
        """
        Fill in any values that are missing in the tree using default
        values from the schema.
        """
        tree = yamlutil.custom_tree_to_tagged_tree(self._tree, self)
        schema.fill_defaults(tree, self)
        self._tree = yamlutil.tagged_tree_to_custom_tree(tree, self)

    def remove_defaults(self):
        """
        Remove any values in the tree that are the same as the default
        values in the schema
        """
        tree = yamlutil.custom_tree_to_tagged_tree(self._tree, self)
        schema.remove_defaults(tree, self)
        self._tree = yamlutil.tagged_tree_to_custom_tree(tree, self)

    def add_history_entry(self, description, software=None):
        """
        Add an entry to the history list.

        Parameters
        ----------
        description : str
            A description of the change.

        software : dict or list of dict
            A description of the software used.  It should not include
            asdf itself, as that is automatically notated in the
            ``asdf_library`` entry.

            Each dict must have the following keys:

            - ``name``: The name of the software
            - ``author``: The author or institution that produced the software
            - ``homepage``: A URI to the homepage of the software
            - ``version``: The version of the software
        """
        if isinstance(software, list):
            software = [Software(x) for x in software]
        elif software is not None:
            software = Software(software)

        time_ = datetime.datetime.fromtimestamp(
            int(os.environ.get("SOURCE_DATE_EPOCH", time.time())), datetime.timezone.utc
        )

        entry = HistoryEntry(
            {
                "description": description,
                "time": time_,
            },
        )

        if software is not None:
            entry["software"] = software

        if self.version >= versioning.NEW_HISTORY_FORMAT_MIN_VERSION:
            if "history" not in self.tree:
                self.tree["history"] = {"entries": []}
            elif "entries" not in self.tree["history"]:
                self.tree["history"]["entries"] = []

            self.tree["history"]["entries"].append(entry)

            try:
                self.validate()
            except Exception:
                self.tree["history"]["entries"].pop()
                raise
        else:
            if "history" not in self.tree:
                self.tree["history"] = []

            self.tree["history"].append(entry)

            try:
                self.validate()
            except Exception:
                self.tree["history"].pop()
                raise

    def get_history_entries(self):
        """
        Get a list of history entries from the file object.

        Returns
        -------
        entries : list
            A list of history entries.
        """

        if "history" not in self.tree:
            return []

        if isinstance(self.tree["history"], list):
            return self.tree["history"]

        if "entries" in self.tree["history"]:
            return self.tree["history"]["entries"]

        return []

    def schema_info(self, key="description", path=None, preserve_list=True):
        """
        Get a nested dictionary of the schema information for a given key, relative to the path.

        This method will only return unambiguous info. If a property is subject to multiple
        subschemas or contains ambiguous entries (multiple titles) no result will be returned
        for that property.

        Parameters
        ----------
        key : str
            The key to look up.
            Default: "description"
        path : str or asdf.search.AsdfSearchResult
            A dot-separated path to the parameter to find the key information on or
            an `asdf.search.AsdfSearchResult` object.
            Default = None (full dictionary).
        preserve_list : bool
            If True, then lists are preserved. Otherwise, they are turned into dicts.
        """
        if isinstance(path, AsdfSearchResult):
            return path.schema_info(
                key,
                preserve_list=preserve_list,
            )

        return node_info.collect_schema_info(
            key,
            path,
            self.tree,
            preserve_list=preserve_list,
            extension_manager=self.extension_manager,
        )

    def info(
        self,
        max_rows=display.DEFAULT_MAX_ROWS,
        max_cols=display.DEFAULT_MAX_COLS,
        show_values=display.DEFAULT_SHOW_VALUES,
    ):
        """
        Print a rendering of this file's tree to stdout.

        Parameters
        ----------
        max_rows : int, tuple, or None, optional
            Maximum number of lines to print.  Nodes that cannot be
            displayed will be elided with a message.
            If int, constrain total number of displayed lines.
            If tuple, constrain lines per node at the depth corresponding \
                to the tuple index.
            If None, display all lines.

        max_cols : int or None, optional
            Maximum length of line to print.  Nodes that cannot
            be fully displayed will be truncated with a message.
            If int, constrain length of displayed lines.
            If None, line length is unconstrained.

        show_values : bool, optional
            Set to False to disable display of primitive values in
            the rendered tree.
        """
        lines = display.render_tree(
            self.tree,
            max_rows=max_rows,
            max_cols=max_cols,
            show_values=show_values,
            identifier="root",
            extension_manager=self.extension_manager,
        )
        print("\n".join(lines))

    def search(self, key=NotSet, type_=NotSet, value=NotSet, filter_=None):
        """
        Search this file's tree.

        Parameters
        ----------
        key : NotSet, str, or any other object
            Search query that selects nodes by dict key or list index.
            If NotSet, the node key is unconstrained.
            If str, the input is searched among keys/indexes as a regular
            expression pattern.
            If any other object, node's key or index must equal the queried key.

        type_ : NotSet, str, or builtins.type
            Search query that selects nodes by type.
            If NotSet, the node type is unconstrained.
            If str, the input is searched among (fully qualified) node type
            names as a regular expression pattern.
            If builtins.type, the node must be an instance of the input.

        value : NotSet, str, or any other object
            Search query that selects nodes by value.
            If NotSet, the node value is unconstrained.
            If str, the input is searched among values as a regular
            expression pattern.
            If any other object, node's value must equal the queried value.

        filter_ : callable
            Callable that filters nodes by arbitrary criteria.
            The callable accepts one or two arguments:

            - the node
            - the node's list index or dict key (optional)

            and returns True to retain the node, or False to remove it from
            the search results.

        Returns
        -------
        asdf.search.AsdfSearchResult
            the result of the search
        """
        result = AsdfSearchResult(["root"], self.tree)
        return result.search(key=key, type_=type_, value=value, filter_=filter_)

    # This function is called from within yamlutil methods to create
    # a context when one isn't explicitly passed in.
    def _create_serialization_context(self, operation=_serialization_context.BlockAccess.NONE):
        return _serialization_context.create(self, operation)


def open_asdf(
    fd,
    uri=None,
    mode=None,
    validate_checksums=False,
    extensions=None,
    ignore_unrecognized_tag=False,
    _force_raw_types=False,
    memmap=False,
    lazy_tree=NotSet,
    lazy_load=True,
    custom_schema=None,
    strict_extension_check=False,
    ignore_missing_extensions=False,
):
    """
    Open an existing ASDF file.

    Parameters
    ----------
    fd : string or file-like object
        May be a string ``file`` or ``http`` URI, or a Python
        file-like object.

    uri : string, optional
        The URI of the file.  Only required if the URI can not be
        automatically determined from ``fd``.

    mode : string, optional
        The mode to open the file in.  Must be ``r`` (default) or
        ``rw``.

    validate_checksums : bool, optional
        If `True`, validate the blocks against their checksums.
        Requires reading the entire file, so disabled by default.

    extensions : object, optional
        Additional extensions to use when reading and writing the file.
        May be an `asdf.extension.Extension` or a `list` of extensions.

    ignore_unrecognized_tag : bool, optional
        When `True`, do not raise warnings for unrecognized tags. Set to
        `False` by default.

    memmap : bool, optional
        When `True`, when reading files, attempt to memmap underlying data
        arrays when possible. Defaults to ``False``.

    lazy_load : bool, optional
        When `True` and the underlying file handle is seekable, data
        arrays will only be loaded lazily: i.e. when they are accessed
        for the first time. In this case the underlying file must stay
        open during the lifetime of the tree. Setting to False causes
        all data arrays to be loaded up front, which means that they
        can be accessed even after the underlying file is closed.
        Note: even if ``lazy_load`` is `False`, ``memmap`` is still taken
        into account.

    lazy_tree : bool, optional
        When `True` the ASDF tree will not be converted to custom objects
        when the file is loaded. Instead, objects will be "lazily" converted
        only when they are accessed. Note that the tree will not contain dict
        and list instances for containers and instead return instances of classes
        defined in `asdf.lazy_nodes`. Since objects are converted when they
        are accessed, traversing the tree (like is done during `AsdfFile.info`
        and `AsdfFile.search`) will result in nodes being converted.

    custom_schema : str, optional
        Path to a custom schema file that will be used for a secondary
        validation pass. This can be used to ensure that particular ASDF
        files follow custom conventions beyond those enforced by the
        specification.

    strict_extension_check : bool, optional
        When `True`, if the given ASDF file contains metadata about the
        extensions used to create it, and if those extensions are not
        installed, opening the file will fail. When `False`, opening a file
        under such conditions will cause only a warning. Defaults to
        `False`.

    ignore_missing_extensions : bool, optional
        When `True`, do not raise warnings when a file is read that
        contains metadata about extensions that are not available. Defaults
        to `False`.

    Returns
    -------
    asdffile : AsdfFile
        The new AsdfFile object.
    """

    if strict_extension_check and ignore_missing_extensions:
        msg = "'strict_extension_check' and 'ignore_missing_extensions' are incompatible options"
        raise ValueError(msg)

    if lazy_tree is NotSet:
        lazy_tree = get_config().lazy_tree

    instance = AsdfFile(
        ignore_unrecognized_tag=ignore_unrecognized_tag,
        memmap=memmap,
        lazy_load=lazy_load,
        custom_schema=custom_schema,
        extensions=extensions,
    )

    with _io.maybe_close(fd, mode, uri) as generic_file:
        file_format_version, comments, tree, blocks = _io.open_asdf(
            generic_file, uri, mode, lazy_load, memmap, validate_checksums
        )

        instance._blocks = blocks
        instance._mode = generic_file.mode
        instance._fd = generic_file
        instance._file_format_version = file_format_version
        instance._comments = comments

        instance.version = _io.find_asdf_version_in_comments(comments, versioning.AsdfVersion("1.0.0"))

        if tree is None:
            # At this point the tree should be tagged, but we want it to be
            # tagged with the core/asdf version appropriate to this file's
            # ASDF Standard version.  We're using custom_tree_to_tagged_tree
            # to select the correct tag for us.
            tree = yamlutil.custom_tree_to_tagged_tree(AsdfObject(), instance)

        # possibly fill defaults
        if instance.version <= versioning.FILL_DEFAULTS_MAX_VERSION and get_config().legacy_fill_schema_defaults:
            schema.fill_defaults(tree, instance, reading=True)

        # validate
        if get_config().validate_on_read:
            instance._validate(tree, reading=True)

        # lazy tree?
        if lazy_tree and not _force_raw_types:
            obj = AsdfObject()
            obj.data = lazy_nodes.AsdfDictNode(tree, weakref.ref(instance))
            tree = obj
        else:
            tree = yamlutil.tagged_tree_to_custom_tree(tree, instance, _force_raw_types)

        # check extensions
        if not (ignore_missing_extensions or _force_raw_types):
            instance._check_extensions(tree, strict=strict_extension_check)

        instance._tree = tree

    return instance



"""
Support for plugins that extend asdf to serialize
additional custom types.
"""

from ._compressor import Compressor
from ._converter import Converter, ConverterProxy
from ._extension import Extension, ExtensionProxy
from ._manager import ExtensionManager, get_cached_extension_manager
from ._manifest import ManifestExtension
from ._serialization_context import SerializationContext
from ._tag import TagDefinition
from ._validator import Validator

__all__ = [
    # New API
    "Compressor",
    "Converter",
    "ConverterProxy",
    "Extension",
    "ExtensionManager",
    "ExtensionProxy",
    "ManifestExtension",
    "SerializationContext",
    "TagDefinition",
    "Validator",
    "get_cached_extension_manager",
]



"""
Commands for searching ASDF trees
"""

import json

import asdf

from .main import Command

__all__ = ["search"]


class Search(Command):
    @classmethod
    def setup_arguments(cls, subparsers):
        parser = subparsers.add_parser(
            "search",
            help="Search an ASDF file.",
            description="Search the contents of an ASDF file and render the search result as text.",
        )

        parser.add_argument("filename", help="ASDF file to search")
        parser.add_argument(
            "--key",
            type=str,
            help="If provided, search the ASDF tree for keys that match the provided regular expression.",
        )
        parser.add_argument(
            "--type",
            type=str,
            help=(
                "If provided, search the ASDF tree for nodes of the provided type. "
                "Must be an importable path or member of builtins. "
                "For arrays either numpy.ndarray or asdf.tags.core.ndarray.NDArrayType can be used."
            ),
        )
        parser.add_argument(
            "--value",
            type=str,
            help=(
                "If provided, search the ASDF tree for nodes that match the provided value. "
                "The provided string will be json decoded (and fall back to the raw string if json decoding fails)."
            ),
        )
        parser.add_argument(
            "--max-rows", type=int, help="Maximum number of lines to print. If not provided all lines will be shown."
        )
        parser.add_argument(
            "--max-cols", type=int, help="Maximum length of line. If not provided lines will have no length limit."
        )

        parser.add_argument(
            "--show-values",
            dest="show_values",
            action="store_true",
            help="Display primitive values in the rendered tree (by default, enabled).",
        )
        parser.add_argument("--no-show-values", dest="show_values", action="store_false")
        parser.set_defaults(show_values=True)

        parser.set_defaults(func=cls.run)

        return parser

    @classmethod
    def run(cls, args):
        search(args.filename, args.key, args.type, args.value, args.max_rows, args.max_cols, args.show_values)


def search(filename, key, type_, value, max_rows, max_cols, show_values):
    query = {}
    if key is not None:
        query["key"] = key
    if type_ is not None:
        if type_ == "numpy.ndarray":
            type_ = "asdf.tags.core.ndarray.NDArrayType"
        if "." not in type_:
            type_ = f"builtins.{type_}"
        module, class_name = type_.rsplit(".", maxsplit=1)
        query["type_"] = getattr(__import__(module, fromlist=[class_name]), class_name)
    if value is not None:
        try:
            value = json.loads(value)
        except json.decoder.JSONDecodeError:
            pass
        query["value"] = value
    with asdf.open(filename) as af:
        result = af.search(**query)
        print(result.format(max_rows, max_cols, show_values))



import warnings
from collections import OrderedDict
from types import GeneratorType

import numpy as np
import yaml

from . import schema, tagged, treeutil, util
from .config import get_config
from .constants import STSCI_SCHEMA_TAG_BASE, YAML_TAG_PREFIX
from .exceptions import AsdfConversionWarning, AsdfSerializationError
from .extension._serialization_context import BlockAccess
from .tags.core import AsdfObject
from .versioning import _YAML_VERSION, _yaml_base_loader

__all__ = ["custom_tree_to_tagged_tree", "tagged_tree_to_custom_tree"]


_yaml_base_dumper = yaml.CSafeDumper if getattr(yaml, "__with_libyaml__", None) else yaml.SafeDumper


YAML_OMAP_TAG = YAML_TAG_PREFIX + "omap"


# ----------------------------------------------------------------------
# Custom loader/dumpers


class AsdfDumper(_yaml_base_dumper):
    """
    A specialized YAML dumper that understands "tagged basic Python
    data types" as implemented in the `tagged` module.
    """

    def __init__(self, *args, **kwargs):
        kwargs["default_flow_style"] = None
        super().__init__(*args, **kwargs)

    def represent_data(self, data):
        node = super().represent_data(data)

        tag_name = getattr(data, "_tag", None)
        if tag_name is not None:
            node.tag = tag_name

        return node


_flow_style_map = {"flow": True, "block": False}


def represent_sequence(dumper, sequence):
    flow_style = _flow_style_map.get(sequence.flow_style, None)
    sequence = sequence.data
    return super(AsdfDumper, dumper).represent_sequence(None, sequence, flow_style)


def represent_mapping(dumper, mapping):
    flow_style = _flow_style_map.get(mapping.flow_style, None)
    node = super(AsdfDumper, dumper).represent_mapping(None, mapping.data, flow_style)

    if mapping.property_order:
        values = node.value
        new_mapping = {}
        for key, val in values:
            new_mapping[key.value] = (key, val)

        new_values = []
        for key in mapping.property_order:
            if key in mapping:
                new_values.append(new_mapping[key])

        property_order = set(mapping.property_order)
        for key, val in values:
            if key.value not in property_order:
                new_values.append((key, val))

        node.value = new_values

    return node


_style_map = {"inline": '"', "folded": ">", "literal": "|"}


def represent_scalar(dumper, value):
    style = _style_map.get(value.style, None)
    return super(AsdfDumper, dumper).represent_scalar(None, value.data, style)


def represent_ordered_mapping(dumper, tag, data):
    # TODO: Again, adjust for preferred flow style, and other stylistic details
    # NOTE: For block style this uses the compact omap notation, but for flow style
    # it does not.

    # TODO: Need to see if I can figure out a mechanism so that classes that
    # use this representer can specify which values should use flow style
    values = []
    node = yaml.SequenceNode(tag, values, flow_style=dumper.default_flow_style)
    if dumper.alias_key is not None:
        dumper.represented_objects[dumper.alias_key] = node
    for key, value in data.items():
        key_item = dumper.represent_data(key)
        value_item = dumper.represent_data(value)
        node_item = yaml.MappingNode(YAML_OMAP_TAG, [(key_item, value_item)], flow_style=False)
        values.append(node_item)
    return node


def represent_ordereddict(dumper, data):
    return represent_ordered_mapping(dumper, YAML_OMAP_TAG, data)


AsdfDumper.add_representer(tagged.TaggedList, represent_sequence)
AsdfDumper.add_representer(tagged.TaggedDict, represent_mapping)
AsdfDumper.add_representer(tagged.TaggedString, represent_scalar)
AsdfDumper.add_representer(OrderedDict, represent_ordereddict)

# ----------------------------------------------------------------------
# Handle numpy scalars


for scalar_type in util._iter_subclasses(np.floating):
    AsdfDumper.add_representer(scalar_type, lambda dumper, data: dumper.represent_float(float(data)))

for scalar_type in util._iter_subclasses(np.integer):
    AsdfDumper.add_representer(scalar_type, lambda dumper, data: dumper.represent_int(int(data)))


def represent_numpy_str(dumper, data):
    # The CSafeDumper implementation will raise an error if it
    # doesn't recognize data as a string.  The Python SafeDumper
    # has no problem with np.str_.
    return dumper.represent_str(str(data))


AsdfDumper.add_representer(np.str_, represent_numpy_str)
AsdfDumper.add_representer(np.bytes_, AsdfDumper.represent_binary)


class _IgnoreCustomTagsLoader(_yaml_base_loader):
    """
    A specialized YAML loader that ignores tags unknown to the
    base (safe) loader. This is used by `asdf.util.load_yaml`
    to read the ASDF tree as "basic" objects, ignoring the
    custom tags.
    """

    def construct_undefined(self, node):
        if isinstance(node, yaml.MappingNode):
            return self.construct_yaml_map(node)
        elif isinstance(node, yaml.SequenceNode):
            return self.construct_yaml_seq(node)
        elif isinstance(node, yaml.ScalarNode):
            return self.construct_scalar(node)
        return super().construct_undefined(node)


# pyyaml will invoke the constructor associated with None when a node's
# tag is not explicitly handled by another constructor.
_IgnoreCustomTagsLoader.add_constructor(None, _IgnoreCustomTagsLoader.construct_undefined)


class AsdfLoader(_yaml_base_loader):
    """
    A specialized YAML loader that can construct "tagged basic Python
    data types" as implemented in the `tagged` module.
    """

    def construct_undefined(self, node):
        if isinstance(node, yaml.MappingNode):
            return self._construct_tagged_mapping(node)

        if isinstance(node, yaml.SequenceNode):
            return self._construct_tagged_sequence(node)

        if isinstance(node, yaml.ScalarNode):
            return self._construct_tagged_scalar(node)

        return super().construct_undefined(node)

    def _construct_tagged_mapping(self, node):
        data = tagged.tag_object(node.tag, {})
        yield data
        data.update(self.construct_mapping(node))

    def _construct_tagged_sequence(self, node):
        data = tagged.tag_object(node.tag, [])
        yield data
        data.extend(self.construct_sequence(node))

    def _construct_tagged_scalar(self, node):
        return tagged.tag_object(node.tag, self.construct_scalar(node))

    # Custom omap deserializer that builds an OrderedDict instead
    # of a list of tuples.  Code is mostly identical to pyyaml's SafeConstructor.
    def construct_yaml_omap(self, node):
        omap = OrderedDict()
        yield omap
        if not isinstance(node, yaml.SequenceNode):
            msg = "while constructing an ordered map"
            raise yaml.constructor.ConstructorError(
                msg,
                node.start_mark,
                f"expected a sequence, but found {node.id}",
                node.start_mark,
            )
        for subnode in node.value:
            if not isinstance(subnode, yaml.MappingNode):
                msg = "while constructing an ordered map"
                raise yaml.constructor.ConstructorError(
                    msg,
                    node.start_mark,
                    f"expected a mapping of length 1, but found {subnode.id}",
                    subnode.start_mark,
                )
            if len(subnode.value) != 1:
                msg = "while constructing an ordered map"
                raise yaml.constructor.ConstructorError(
                    msg,
                    node.start_mark,
                    f"expected a single mapping item, but found {len(subnode.value)} items",
                    subnode.start_mark,
                )
            key_node, value_node = subnode.value[0]
            key = self.construct_object(key_node)
            value = self.construct_object(value_node)
            omap[key] = value


# pyyaml will invoke the constructor associated with None when a node's
# tag is not explicitly handled by another constructor.
AsdfLoader.add_constructor(None, AsdfLoader.construct_undefined)
AsdfLoader.add_constructor(YAML_TAG_PREFIX + "omap", AsdfLoader.construct_yaml_omap)


def custom_tree_to_tagged_tree(tree, ctx, _serialization_context=None):
    """
    Convert a tree, possibly containing custom data types that aren't
    directly representable in YAML, to a tree of basic data types,
    annotated with tags.
    """
    if _serialization_context is None:
        _serialization_context = ctx._create_serialization_context(BlockAccess.WRITE)

    extension_manager = _serialization_context.extension_manager

    def _convert_obj(obj, converter):
        tag = converter.select_tag(obj, _serialization_context)
        # if select_tag returns None, converter.to_yaml_tree should return a new
        # object which will be handled by a different converter
        converters_used = set()
        while tag is None:
            converters_used.add(converter)
            obj = converter.to_yaml_tree(obj, tag, _serialization_context)
            try:
                converter = extension_manager.get_converter_for_type(type(obj))
            except KeyError:
                # no converter supports this type, return it as-is
                yield obj
                return
            if converter in converters_used:
                msg = "Conversion cycle detected"
                raise TypeError(msg)
            tag = converter.select_tag(obj, _serialization_context)
        _serialization_context.assign_object(obj)
        node = converter.to_yaml_tree(obj, tag, _serialization_context)
        _serialization_context.assign_blocks()

        if isinstance(node, GeneratorType):
            generator = node
            node = next(generator)
        else:
            generator = None

        if isinstance(node, dict):
            tagged_node = tagged.TaggedDict(node, tag)
        elif isinstance(node, list):
            tagged_node = tagged.TaggedList(node, tag)
        elif isinstance(node, str):
            tagged_node = tagged.TaggedString(node)
            tagged_node._tag = tag
        else:
            msg = f"Converter returned illegal node type: {util.get_class_name(node)}"
            raise TypeError(msg)

        _serialization_context._mark_extension_used(converter.extension)

        yield tagged_node
        if generator is not None:
            yield from generator

    converters_cache = {}

    def _walker(obj):
        typ = type(obj)
        if typ in converters_cache:
            return converters_cache[typ](obj)
        if extension_manager.handles_type(typ):
            converter = extension_manager.get_converter_for_type(typ)
            converters_cache[typ] = lambda obj, _converter=converter: _convert_obj(obj, _converter)
            return _convert_obj(obj, converter)

        converters_cache[typ] = lambda obj: obj
        return obj

    return treeutil.walk_and_modify(
        tree,
        _walker,
        # Walk the tree in preorder, so that extensions can return
        # container nodes with unserialized children.
        postorder=False,
        _context=ctx._tree_modification_context,
    )


def tagged_tree_to_custom_tree(tree, ctx, force_raw_types=False, _serialization_context=None):
    """
    Convert a tree containing only basic data types, annotated with
    tags, to a tree containing custom data types.
    """
    if _serialization_context is None:
        _serialization_context = ctx._create_serialization_context(BlockAccess.READ)

    extension_manager = _serialization_context.extension_manager
    cfg = get_config()

    def _walker(node):
        if force_raw_types:
            return node

        tag = getattr(node, "_tag", None)
        if tag is None:
            return node

        if extension_manager.handles_tag(tag):
            converter = extension_manager.get_converter_for_tag(tag)
            try:
                obj = converter.from_yaml_tree(node.data, tag, _serialization_context)
            except Exception as err:
                if cfg.warn_on_failed_conversion:
                    warnings.warn(f"A node failed to convert with: {err}", AsdfConversionWarning)
                    obj = node
                else:
                    raise
            _serialization_context.assign_object(obj)
            _serialization_context.assign_blocks()
            _serialization_context._mark_extension_used(converter.extension)

            if isinstance(obj, GeneratorType) and cfg.warn_on_failed_conversion:
                # wrap the generator to catch any errors
                def wrapped_generator(generator, node):
                    try:
                        yield from generator
                    except Exception as err:
                        warnings.warn(f"A node failed to convert with: {err}", AsdfConversionWarning)
                        yield node

                return wrapped_generator(obj, node)
            return obj

        if not ctx._ignore_unrecognized_tag:
            warnings.warn(
                f"{tag} is not recognized, converting to raw Python data structure",
                AsdfConversionWarning,
            )
        return node

    return treeutil.walk_and_modify(
        tree,
        _walker,
        # Walk the tree in postorder, so that extensions receive
        # container nodes with children already deserialized.
        postorder=True,
        _context=ctx._tree_modification_context,
    )


def load_tree(stream):
    """
    Load YAML, returning a tree of objects.

    Parameters
    ----------
    stream : readable file-like object
        Stream containing the raw YAML content.
    """
    # The following call to yaml.load is safe because we're
    # using a loader that inherits from pyyaml's SafeLoader.
    return yaml.load(stream, Loader=AsdfLoader)  # noqa: S506


def dump_tree(tree, fd, ctx, tree_finalizer=None, _serialization_context=None):
    """
    Dump a tree of objects, possibly containing custom types, to YAML.

    Parameters
    ----------
    tree : object
        Tree of objects, possibly containing custom data types.

    fd : asdf.generic_io.GenericFile
        A file object to dump the serialized YAML to.

    ctx : Context
        The writing context.

    tree_finalizer : callable, optional
        Callback that receives the tagged tree before it is validated
        and defaults are removed.  `asdf.AsdfFile` uses this to update
        extension metadata on the tagged tree before it is fully
        serialized to YAML.
    """
    # The _serialization_context parameter allows AsdfFile to track
    # what extensions were used when converting the tree's custom
    # types.  In 3.0, it will be passed as the `ctx` instead of the
    # AsdfFile itself.
    if type(tree) is not AsdfObject:
        msg = "Root node of ASDF tree must be of type AsdfObject"
        raise TypeError(msg)

    tags = {"!": STSCI_SCHEMA_TAG_BASE + "/"}
    tree = custom_tree_to_tagged_tree(tree, ctx, _serialization_context=_serialization_context)
    if tree_finalizer is not None:
        tree_finalizer(tree)
    schema.validate(tree, ctx)

    # add yaml %TAG definitions from extensions
    if _serialization_context:
        for ext in _serialization_context._extensions_used:
            for key, val in ext.yaml_tag_handles.items():
                if key not in tags:
                    tags[key] = val

    try:
        yaml.dump_all(
            [tree],
            stream=fd,
            Dumper=AsdfDumper,
            explicit_start=True,
            explicit_end=True,
            version=_YAML_VERSION,
            allow_unicode=True,
            encoding="utf-8",
            tags=tags,
        )
    except yaml.representer.RepresenterError as err:
        if len(err.args) < 2:
            raise err
        # inspect the exception arguments to determine what object failed
        obj = err.args[1]
        msg = (
            f"Object of type[{type(obj)}] is not serializable by asdf. "
            "Please convert the object to a supported type or implement "
            "a Converter for this type to allow the tree to be serialized."
        )
        raise AsdfSerializationError(msg, obj) from err



"""
Implementation of command for reporting information about installed extensions.
"""

from asdf._entry_points import get_extensions

from .main import Command

__all__ = ["find_extensions"]


class QueryExtension(Command):  # pragma: no cover
    """This class is the plugin implementation for the asdftool runner."""

    @classmethod
    def setup_arguments(cls, subparsers):
        parser = subparsers.add_parser(
            "extensions",
            help="Show information about installed extensions",
            description="""Reports information about installed ASDF extensions""",
        )

        display_group = parser.add_mutually_exclusive_group()
        display_group.add_argument(
            "-s",
            "--summary",
            action="store_true",
            help="Display only the installed extensions themselves",
        )
        display_group.add_argument(
            "-t",
            "--tags-only",
            action="store_true",
            help="Display tags from installed extensions, but no other information",
        )

        parser.set_defaults(func=cls.run)
        return parser

    @classmethod
    def run(cls, args):
        return find_extensions(args.summary, args.tags_only)


def _format_extension(ext):
    uri = "(none)" if ext.extension_uri is None else f"'{ext.extension_uri}'"

    return f"Extension URI: {uri} package: {ext.package_name} ({ext.package_version}) class: {ext.class_name}"


def _format_type_name(type_):
    if isinstance(type_, str):
        return type_

    return f"{type_.__module__}.{type_.__name__}"


def _print_extension_details(ext, tags_only):
    tag_uris = [t.tag_uri for t in ext.tags]
    for typ in ext.types:
        if isinstance(typ.name, list):
            for name in typ.name:
                tag_uris.append(typ.make_yaml_tag(name))
        elif typ.name is not None:
            tag_uris.append(typ.make_yaml_tag(typ.name))

    if len(tag_uris) > 0:
        print("tags:")
        for tag_uri in sorted(tag_uris):
            print(f"  - {tag_uri}")

    if not tags_only:
        types = []
        for converter in ext.converters:
            for typ in converter.types:
                types.append(typ)
        for typ in ext.types:
            types.extend(typ.types)

        if len(types) > 0:
            print("types:")
            for typ in sorted(types, key=_format_type_name):
                print(f"  - {_format_type_name(typ)}")


def find_extensions(summary, tags_only):
    for ext in get_extensions():
        print(_format_extension(ext))
        if not summary:
            _print_extension_details(ext, tags_only)
            print()



from asdf.extension import Validator
from asdf.tags.core.ndarray import validate_datatype, validate_max_ndim, validate_ndim


class NdimValidator(Validator):
    schema_property = "ndim"
    tags = ["tag:stsci.edu:asdf/core/ndarray-*"]

    def validate(self, expected_ndim, node, schema):
        yield from validate_ndim(None, expected_ndim, node, schema)


class MaxNdimValidator(Validator):
    schema_property = "max_ndim"
    tags = ["tag:stsci.edu:asdf/core/ndarray-*"]

    def validate(self, max_ndim, node, schema):
        yield from validate_max_ndim(None, max_ndim, node, schema)


class DatatypeValidator(Validator):
    schema_property = "datatype"
    tags = ["tag:stsci.edu:asdf/core/ndarray-*"]

    def validate(self, expected_datatype, node, schema):
        yield from validate_datatype(None, expected_datatype, node, schema)



import numpy as np
import pytest

import asdf
from asdf import AsdfFile
from asdf._commands import main
from asdf._tests._helpers import assert_tree_match


def _test_defragment(tmp_path, codec):
    x = np.arange(0, 1000, dtype=float)

    tree = {
        "science_data": x,
        "subset": x[3:-3],
        "skipping": x[::2],
        "not_shared": np.arange(100, 0, -1, dtype=np.uint8),
    }

    path = tmp_path / "original.asdf"
    out_path = tmp_path / "original.defragment.asdf"
    ff = AsdfFile(tree)
    ff.write_to(path)
    with asdf.open(path) as af:
        assert len(af._blocks.blocks) == 2

    with pytest.warns(UserWarning, match="defragment is deprecated"):
        result = main.main_from_args(["defragment", str(path), "-o", str(out_path), "-c", codec])

    assert result == 0

    files = [p.name for p in tmp_path.iterdir()]

    assert "original.asdf" in files
    assert "original.defragment.asdf" in files

    original_size = (tmp_path / "original.asdf").stat().st_size
    defragment_size = (tmp_path / "original.defragment.asdf").stat().st_size
    assert original_size > defragment_size

    with asdf.open(tmp_path / "original.defragment.asdf") as ff:
        assert_tree_match(ff.tree, tree)
        assert len(ff._blocks.blocks) == 2


def test_defragment_zlib(tmp_path):
    _test_defragment(tmp_path, "zlib")


def test_defragment_bzp2(tmp_path):
    _test_defragment(tmp_path, "bzp2")


def test_defragment_lz4(tmp_path):
    pytest.importorskip("lz4")
    _test_defragment(tmp_path, "lz4")



"""
Utilities for searching ASDF trees.
"""

import builtins
import inspect
import re
import typing

from ._display import DEFAULT_MAX_COLS, DEFAULT_MAX_ROWS, DEFAULT_SHOW_VALUES, render_tree
from ._node_info import collect_schema_info
from .treeutil import get_children, is_container
from .util import NotSet

__all__ = ["AsdfSearchResult"]


class AsdfSearchResult:
    """
    Result of a call to AsdfFile.search.
    """

    def __init__(
        self,
        identifiers,
        node,
        filters=None,
        parent_node=None,
        max_rows=DEFAULT_MAX_ROWS,
        max_cols=DEFAULT_MAX_COLS,
        show_values=DEFAULT_SHOW_VALUES,
    ):
        self._identifiers = identifiers
        self._node = node
        self._filters = [] if filters is None else filters
        self._parent_node = parent_node
        self._max_rows = max_rows
        self._max_cols = max_cols
        self._show_values = show_values

    def format(self, max_rows=NotSet, max_cols=NotSet, show_values=NotSet):
        """
        Change formatting parameters of the rendered tree.

        Parameters
        ----------
        max_rows : int, tuple, None, or NotSet, optional
            Maximum number of lines to print.  Nodes that cannot be
            displayed will be elided with a message.
            If int, constrain total number of displayed lines.
            If tuple, constrain lines per node at the depth corresponding \
                to the tuple index.
            If None, display all lines.
            If NotSet, retain existing value.

        max_cols : int, None or NotSet, optional
            Maximum length of line to print.  Nodes that cannot
            be fully displayed will be truncated with a message.
            If int, constrain length of displayed lines.
            If None, line length is unconstrained.
            If NotSet, retain existing value.

        show_values : bool or NotSet, optional
            Set to False to disable display of primitive values in
            the rendered tree.
            Set to NotSet to retain existign value.

        Returns
        -------
        AsdfSearchResult
            the reformatted search result
        """
        if max_rows is NotSet:
            max_rows = self._max_rows

        if max_cols is NotSet:
            max_cols = self._max_cols

        if show_values is NotSet:
            show_values = self._show_values

        return AsdfSearchResult(
            self._identifiers,
            self._node,
            filters=self._filters,
            parent_node=self._parent_node,
            max_rows=max_rows,
            max_cols=max_cols,
            show_values=show_values,
        )

    def _maybe_compile_pattern(self, query):
        if isinstance(query, str):
            return re.compile(query)

        return query

    def _safe_equals(self, a, b):
        try:
            result = a == b

        except Exception:
            return False

        if isinstance(result, bool):
            return result

        return False

    def _get_fully_qualified_type(self, value):
        value_type = type(value)
        if value_type.__module__ == "builtins":
            return value_type.__name__

        return ".".join([value_type.__module__, value_type.__name__])

    def search(self, key=NotSet, type_=NotSet, value=NotSet, filter_=None):
        """
        Further narrow the search.

        Parameters
        ----------
        key : NotSet, str, or any other object
            Search query that selects nodes by dict key or list index.
            If NotSet, the node key is unconstrained.
            If str, the input is searched among keys/indexes as a regular
            expression pattern.
            If any other object, node's key or index must equal the queried key.

        type_ : NotSet, str, or builtins.type
            Search query that selects nodes by type.
            If NotSet, the node type is unconstrained.
            If str, the input is searched among (fully qualified) node type
            names as a regular expression pattern.
            If builtins.type, the node must be an instance of the input.

        value : NotSet, str, or any other object
            Search query that selects nodes by value.
            If NotSet, the node value is unconstrained.
            If str, the input is searched among values as a regular
            expression pattern.
            If any other object, node's value must equal the queried value.

        filter_ : callable
            Callable that filters nodes by arbitrary criteria.
            The callable accepts one or two arguments:

            - the node
            - the node's list index or dict key (optional)

            and returns True to retain the node, or False to remove it from
            the search results.

        Returns
        -------
        AsdfSearchResult
            the subsequent search result
        """
        if not (isinstance(type_, (str, typing.Pattern, builtins.type)) or type_ is NotSet):
            msg = "type must be NotSet, str, regular expression, or instance of builtins.type"
            raise TypeError(msg)

        # value and key arguments can be anything, but pattern and str have special behavior

        key = self._maybe_compile_pattern(key)
        type_ = self._maybe_compile_pattern(type_)
        value = self._maybe_compile_pattern(value)

        filter_ = _wrap_filter(filter_)

        def _filter(node, identifier):
            if isinstance(key, typing.Pattern):
                if key.search(str(identifier)) is None:
                    return False

            elif key is not NotSet and not self._safe_equals(identifier, key):
                return False

            if isinstance(type_, typing.Pattern):
                fully_qualified_node_type = self._get_fully_qualified_type(node)
                if type_.search(fully_qualified_node_type) is None:
                    return False

            elif isinstance(type_, builtins.type) and not isinstance(node, type_):
                return False

            if isinstance(value, typing.Pattern):
                if is_container(node):
                    # The string representation of a container object tends to
                    # include the child object values, but that's probably not
                    # what searchers want.
                    return False

                if value.search(str(node)) is None:
                    return False

            elif value is not NotSet and not self._safe_equals(node, value):
                return False

            if filter_ is not None and not filter_(node, identifier):
                return False

            return True

        return AsdfSearchResult(
            self._identifiers,
            self._node,
            filters=[*self._filters, _filter],
            parent_node=self._parent_node,
            max_rows=self._max_rows,
            max_cols=self._max_cols,
            show_values=self._show_values,
        )

    def replace(self, value):
        """
        Assign a new value in place of all leaf nodes in the
        search results.

        Parameters
        ----------
        value : object
        """
        results = []

        def _callback(identifiers, parent, node, children):
            if all(f(node, identifiers[-1]) for f in self._filters):
                results.append((identifiers[-1], parent))

        _walk_tree_breadth_first(self._identifiers, self._node, _callback)

        for identifier, parent in results:
            parent[identifier] = value

    @property
    def node(self):
        """
        Retrieve the leaf node of a tree with one search result.

        Returns
        -------
        object
            the single node of the search result
        """

        results = self.nodes
        if len(results) == 0:
            return None

        if len(results) == 1:
            return results[0]

        msg = "More than one result"
        raise RuntimeError(msg)

    @property
    def path(self):
        """
        Retrieve the path to the leaf node of a tree with one search result.

        Returns
        -------
        str
            the path to the searched node
        """
        results = self.paths
        if len(results) == 0:
            return None

        if len(results) == 1:
            return results[0]

        msg = "More than one result"
        raise RuntimeError(msg)

    @property
    def nodes(self):
        """
        Retrieve all leaf nodes in the search results.

        Returns
        -------
        list of object
            every node in the search results (breadth-first order)
        """
        results = []

        def _callback(identifiers, parent, node, children):
            if all(f(node, identifiers[-1]) for f in self._filters):
                results.append(node)

        _walk_tree_breadth_first(self._identifiers, self._node, _callback)
        return results

    @property
    def paths(self):
        """
        Retrieve the paths to all leaf nodes in the search results.

        Returns
        -------
        list of str
            the path to every node in the search results
        """
        results = []

        def _callback(identifiers, parent, node, children):
            if all(f(node, identifiers[-1]) for f in self._filters):
                results.append(_build_path(identifiers))

        _walk_tree_breadth_first(self._identifiers, self._node, _callback)
        return results

    def __repr__(self):
        lines = render_tree(
            self._node,
            max_rows=self._max_rows,
            max_cols=self._max_cols,
            show_values=self._show_values,
            filters=self._filters,
            identifier=self._identifiers[-1],
        )

        if len(lines) == 0:
            return "No results found."

        return "\n".join(lines)

    def schema_info(self, key="description", preserve_list=True):
        """
        Get a nested dictionary of the schema information for a given key, relative to this search result.

        Parameters
        ----------
        key : str
            The key to look up.
            Default: "description"
        preserve_list : bool
            If True, then lists are preserved. Otherwise, they are turned into dicts.
        """
        return collect_schema_info(
            key,
            None,
            self._node,
            filters=self._filters,
            preserve_list=preserve_list,
        )

    def __getitem__(self, key):
        if isinstance(self._node, (dict, list, tuple)) or hasattr(self._node, "__asdf_traverse__"):
            child = self._node.__asdf_traverse__()[key] if hasattr(self._node, "__asdf_traverse__") else self._node[key]
        else:
            msg = "This node cannot be indexed"
            raise TypeError(msg)

        return AsdfSearchResult(
            [*self._identifiers, key],
            child,
            filters=self._filters,
            parent_node=self._node,
            max_rows=self._max_rows,
            max_cols=self._max_cols,
            show_values=self._show_values,
        )


def _walk_tree_breadth_first(root_identifiers, root_node, callback):
    """
    Walk the tree in breadth-first order (useful for prioritizing
    lower-depth nodes).
    """
    current_nodes = [(root_identifiers, None, root_node)]
    seen = set()
    while True:
        next_nodes = []

        for identifiers, parent, node in current_nodes:
            if (isinstance(node, (dict, list, tuple)) or hasattr(node, "__asdf_traverse__")) and id(node) in seen:
                continue
            tnode = node.__asdf_traverse__() if hasattr(node, "__asdf_traverse__") else node
            children = get_children(tnode)
            callback(identifiers, parent, node, [c for _, c in children])
            next_nodes.extend([([*identifiers, i], node, c) for i, c in children])
            seen.add(id(node))

        if len(next_nodes) == 0:
            break

        current_nodes = next_nodes


def _build_path(identifiers):
    """
    Generate the Python code needed to extract the identified node.
    """
    if len(identifiers) == 0:
        return ""

    return identifiers[0] + "".join(f"[{i!r}]" for i in identifiers[1:])


def _wrap_filter(filter_):
    """
    Ensure that filter callable accepts the expected number of arguments.
    """
    if filter_ is None:
        return None

    arity = len(inspect.signature(filter_).parameters)
    if arity == 1:
        return lambda n, i: filter_(n)

    if arity == 2:
        return filter_

    msg = "filter must accept 1 or 2 arguments"
    raise ValueError(msg)



import enum

from asdf._block.key import Key as BlockKey
from asdf._helpers import validate_version
from asdf.extension._extension import ExtensionProxy


class SerializationContext:
    """
    Container for parameters of the current (de)serialization.

    This class should not be instantiated directly and instead
    will be created by the AsdfFile object and provided to extension
    classes (like Converters) via method arguments.
    """

    def __init__(self, version, extension_manager, url, blocks):
        self._version = validate_version(version)
        self._extension_manager = extension_manager
        self._url = url
        self._blocks = blocks
        self._obj = None

        self.__extensions_used = set()

    @property
    def url(self):
        """
        The URL (if any) of the file being read or written.

        Used to compute relative locations of external files referenced by this
        ASDF file. The URL will not exist in some cases (e.g. when the file is
        written to an `io.BytesIO`).

        Returns
        -------
        str or None
        """
        return self._url

    @property
    def version(self):
        """
        Get the ASDF Standard version.

        Returns
        -------
        str
        """
        return self._version

    @property
    def extension_manager(self):
        """
        Get the ExtensionManager for enabled extensions.

        Returns
        -------
        asdf.extension.ExtensionManager
        """
        return self._extension_manager

    def _mark_extension_used(self, extension):
        """
        Note that an extension was used when reading or writing the file.

        Parameters
        ----------
        extension : asdf.extension.Extension
        """
        self.__extensions_used.add(ExtensionProxy.maybe_wrap(extension))

    @property
    def _extensions_used(self):
        """
        Get the set of extensions that were used when reading or writing the file.

        Returns
        -------
        set of asdf.extension.Extension
        """
        return self.__extensions_used

    def get_block_data_callback(self, index, key=None):
        """
        Generate a callable that when called will read data
        from an ASDF block at the provided index.

        Parameters
        ----------
        index : int
            Index of ASDF block.

        key : BlockKey, optional
            BlockKey generated using self.generate_block_key. Only
            needed for a Converter that uses multiple blocks.

        Returns
        -------
        callback : callable
            A callable that when called (with no arguments) returns
            the block data as a one dimensional array of uint8
        """
        raise NotImplementedError("abstract")

    def find_available_block_index(self, data_callback, key=None):
        """
        Find the index of an available ASDF block to write data.

        This is typically used inside asdf.extension.Converter.to_yaml_tree.

        Parameters
        ----------
        data_callback: callable
            Callable that when called will return data (ndarray) that will
            be written to a block.

        key : BlockKey, optional
            BlockKey generated using self.generate_block_key. Only
            needed for a Converter that uses multiple blocks.

        Returns
        -------
        block_index: int
            Index of the ASDF block where data returned from
            data_callback will be written.
        """
        raise NotImplementedError("abstract")

    def generate_block_key(self):
        """
        Generate a BlockKey used for Converters that wish to use
        multiple blocks

        Returns
        -------
        key : BlockKey
            A hashable object that will be associated with the
            serialized/deserialized object and can be used to
            access multiple blocks within a Converter
        """
        raise NotImplementedError("abstract")

    def assign_object(self, obj):
        self._obj = obj

    def assign_blocks(self):
        pass

    def set_array_storage(self, arr, array_storage):
        """
        Set the block type to use for the given array data.

        Parameters
        ----------
        arr : numpy.ndarray
            The array to set.  If multiple views of the array are in
            the tree, only the most recent block type setting will be
            used, since all views share a single block.

        array_storage : str
            Must be one of:

            - ``internal``: The default.  The array data will be
              stored in a binary block in the same ASDF file.

            - ``external``: Store the data in a binary block in a
              separate ASDF file.

            - ``inline``: Store the data as YAML inline in the tree.
        """
        self._blocks._set_array_storage(arr, array_storage)

    def get_array_storage(self, arr):
        """
        Get the block type for the given array data.

        Parameters
        ----------
        arr : numpy.ndarray
        """
        return self._blocks._get_array_storage(arr)

    def set_array_compression(self, arr, compression, **compression_kwargs):
        """
        Set the compression to use for the given array data.

        Parameters
        ----------
        arr : numpy.ndarray
            The array to set.  If multiple views of the array are in
            the tree, only the most recent compression setting will be
            used, since all views share a single block.

        compression : str or None
            Must be one of:

            - ``''`` or `None`: no compression

            - ``zlib``: Use zlib compression

            - ``bzp2``: Use bzip2 compression

            - ``lz4``: Use lz4 compression

            - ``input``: Use the same compression as in the file read.
              If there is no prior file, acts as None.

        """
        self._blocks._set_array_compression(arr, compression, **compression_kwargs)

    def get_array_compression(self, arr):
        """
        Get the compression type for the given array data.

        Parameters
        ----------
        arr : numpy.ndarray

        Returns
        -------
        compression : str or None
        """
        return self._blocks._get_array_compression(arr)

    def get_array_compression_kwargs(self, arr):
        """ """
        return self._blocks._get_array_compression_kwargs(arr)

    def set_array_save_base(self, arr, save_base):
        """
        Set the ``save_base`` option for ``arr``. When ``arr`` is
        written to a file, if ``save_base`` is ``True`` the base array
        for ``arr`` will be saved.

        Note that similar to other array options this setting is linked
        to the base array if ``arr`` is a view.

        Parameters
        ----------
        arr : numpy.ndarray

        save_base : bool or None
            if ``None`` the ``default_array_save_base`` value from asdf
            config will be used
        """
        self._blocks._set_array_save_base(arr, save_base)

    def get_array_save_base(self, arr):
        """
        Returns the ``save_base`` option for ``arr``. When ``arr`` is
        written to a file, if ``save_base`` is ``True`` the base array
        for ``arr`` will be saved.

        Parameters
        ----------
        arr : numpy.ndarray

        Returns
        -------
        save_base : bool
        """
        return self._blocks._get_array_save_base(arr)


class ReadBlocksContext(SerializationContext):
    """
    Perform deserialization (reading) with a `SerializationContext`.

    To allow for block access, `ReadBlocksContext` implements:
        - `SerializationContext.generate_block_key`
        - `SerializationContext.get_block_data_callback`
    and tracks which blocks (and keys) are accessed, assigning them
    to the deserialized object after `assign_object` and
    `assign_blocks` are called.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.assign_object(None)

    def assign_object(self, obj):
        super().assign_object(obj)
        if obj is None:
            self._cb = None
            self._keys_to_assign = {}

    def assign_blocks(self):
        super().assign_blocks()
        if self._cb is not None:
            self._blocks._data_callbacks.assign_object(self._obj, self._cb)
        for key, cb in self._keys_to_assign.items():
            if cb is None:
                msg = "Converter generated a key that was never used"
                raise OSError(msg)
            # now that we have an object, make the key valid
            key._assign_object(self._obj)

            # assign the key to the callback
            self._blocks._data_callbacks.assign_object(key, cb)
        # now that we've assigned blocks, remove the reference to the
        # assigned object
        self.assign_object(None)

    def get_block_data_callback(self, index, key=None):
        if key is None:
            if self._cb is not None:
                # this operation has already accessed a block without using
                # a key so check if the same index was accessed
                if self._cb._index == index:
                    return self._cb
                msg = "Converters accessing >1 block must provide a key for each block"
                raise OSError(msg)
            self._cb = self._blocks._get_data_callback(index)
            return self._cb

        if self._keys_to_assign.get(key, None) is not None:
            return self._keys_to_assign[key]

        cb = self._blocks._get_data_callback(index)
        # mark this as a key to later assign
        self._keys_to_assign[key] = cb
        return cb

    def generate_block_key(self):
        key = BlockKey()
        self._keys_to_assign[key] = None
        return key


class WriteBlocksContext(SerializationContext):
    """
    Perform serialization (writing) with a `SerializationContext`.

    To allow for block access, `WriteBlocksContext` implements:
        - `SerializationContext.generate_block_key`
        - `SerializationContext.find_available_block_index`
    and assigns any accessed blocks (and keys) to the object
    being serialized.
    """

    def find_available_block_index(self, data_callback, key=None):
        if key is None:
            key = self._obj
        return self._blocks.make_write_block(data_callback, None, key)

    def generate_block_key(self):
        return BlockKey(self._obj)


class BlockAccess(enum.Enum):
    """
    Block access enumerated values that define
    how a SerializationContext can access ASDF blocks.
    """

    NONE = SerializationContext
    WRITE = WriteBlocksContext
    READ = ReadBlocksContext


def create(asdf_file, block_access=BlockAccess.NONE):
    """
    Create a SerializationContext instance (or subclass) using
    an AsdfFile instance, asdf_file.

    Parameters
    ----------
    asdf_file : asdf.AsdfFile

    block_access : BlockAccess, optional
        Defaults to BlockAccess.NONE
    """
    return block_access.value(asdf_file.version_string, asdf_file.extension_manager, asdf_file.uri, asdf_file._blocks)



import numpy as np
import pytest

import asdf


@pytest.fixture
def software_asdf_file():
    return asdf.AsdfFile({"obj": asdf.tags.core.Software(name="foo", version="0.0.0")})


@pytest.fixture
def ndarray_asdf_file():
    return asdf.AsdfFile({"obj": np.ndarray([1])})


@pytest.fixture(params=["software_asdf_file", "ndarray_asdf_file"])
def asdf_file(request):
    return request.getfixturevalue(request.param)


def test_validate(asdf_file, benchmark):
    # first validate outside the benchmark to incur
    # extension loading, schema caching and other one-time costs
    asdf_file.validate()
    benchmark(asdf_file.validate)



import pytest

from asdf._commands import find_extensions
from asdf.versioning import supported_versions


@pytest.mark.parametrize("summary", [True, False])
@pytest.mark.parametrize("tags_only", [True, False])
def test_parameter_combinations(summary, tags_only):
    # Just confirming no errors:
    find_extensions(summary, tags_only)


@pytest.mark.parametrize("standard_version", supported_versions)
def test_builtin_extension_included(capsys, standard_version):
    find_extensions(True, False)
    captured = capsys.readouterr()
    assert f"core-{standard_version}" in captured.out



"""
Methods for getting and setting asdf global configuration
options.
"""

import collections
import copy
import threading
from contextlib import contextmanager

from . import _entry_points, util, versioning
from ._helpers import validate_version
from .extension import ExtensionProxy
from .resource import ResourceManager, ResourceMappingProxy

__all__ = ["AsdfConfig", "config_context", "get_config"]


DEFAULT_VALIDATE_ON_READ = True
DEFAULT_DEFAULT_VERSION = str(versioning.default_version)
DEFAULT_LEGACY_FILL_SCHEMA_DEFAULTS = True
DEFAULT_IO_BLOCK_SIZE = -1  # auto
DEFAULT_ARRAY_INLINE_THRESHOLD = None
DEFAULT_ALL_ARRAY_STORAGE = None
DEFAULT_ALL_ARRAY_COMPRESSION = "input"
DEFAULT_ALL_ARRAY_COMPRESSION_KWARGS = None
DEFAULT_DEFAULT_ARRAY_SAVE_BASE = True
DEFAULT_LAZY_TREE = False
DEFAULT_WARN_ON_FAILED_CONVERSION = False


class AsdfConfig:
    """
    Container for ASDF configuration options.  Users are not intended to
    construct this object directly; instead, use the `asdf.get_config` and
    `asdf.config_context` module methods.
    """

    def __init__(self):
        self._resource_mappings = None
        self._resource_manager = None
        self._extensions = None
        self._validate_on_read = DEFAULT_VALIDATE_ON_READ
        self._default_version = DEFAULT_DEFAULT_VERSION
        self._legacy_fill_schema_defaults = DEFAULT_LEGACY_FILL_SCHEMA_DEFAULTS
        self._io_block_size = DEFAULT_IO_BLOCK_SIZE
        self._array_inline_threshold = DEFAULT_ARRAY_INLINE_THRESHOLD
        self._all_array_storage = DEFAULT_ALL_ARRAY_STORAGE
        self._all_array_compression = DEFAULT_ALL_ARRAY_COMPRESSION
        self._all_array_compression_kwargs = DEFAULT_ALL_ARRAY_COMPRESSION_KWARGS
        self._default_array_save_base = DEFAULT_DEFAULT_ARRAY_SAVE_BASE
        self._lazy_tree = DEFAULT_LAZY_TREE
        self._warn_on_failed_conversion = DEFAULT_WARN_ON_FAILED_CONVERSION

        self._lock = threading.RLock()

    @property
    def resource_mappings(self):
        """
        Get the list of registered resource mapping instances.  Unless
        overridden by user configuration, this list contains every mapping
        registered with an entry point.

        Returns
        -------
        list of asdf.resource.ResourceMappingProxy
        """
        if self._resource_mappings is None:
            with self._lock:
                if self._resource_mappings is None:
                    self._resource_mappings = _entry_points.get_resource_mappings()
        return self._resource_mappings

    def add_resource_mapping(self, mapping):
        """
        Register a new resource mapping.  The new mapping will
        take precedence over all previously registered mappings.

        Parameters
        ----------
        mapping : collections.abc.Mapping
            Map of `str` resource URI to `bytes` content
        """
        with self._lock:
            mapping = ResourceMappingProxy.maybe_wrap(mapping)

            # Insert at the beginning of the list so that
            # ResourceManager uses the new mapping first.
            resource_mappings = [mapping] + [r for r in self.resource_mappings if r != mapping]
            self._resource_mappings = resource_mappings
            self._resource_manager = None

    def remove_resource_mapping(self, mapping=None, *, package=None):
        """
        Remove a registered resource mapping.

        Parameters
        ----------
        mapping : collections.abc.Mapping, optional
            Mapping to remove.
        package : str, optional
            Remove only extensions provided by this package.  If the ``mapping``
            argument is omitted, then all mappings from this package will
            be removed.
        """
        if mapping is None and package is None:
            msg = "Must specify at least one of mapping or package"
            raise ValueError(msg)

        if mapping is not None:
            mapping = ResourceMappingProxy.maybe_wrap(mapping)

        def _remove_condition(m):
            result = True
            if mapping is not None:
                result = result and m == mapping
            if package is not None:
                result = result and m.package_name == package

            return result

        with self._lock:
            self._resource_mappings = [m for m in self.resource_mappings if not _remove_condition(m)]
            self._resource_manager = None

    def reset_resources(self):
        """
        Reset registered resource mappings to the default list
        provided as entry points.
        """
        with self._lock:
            self._resource_mappings = None
            self._resource_manager = None

    @property
    def resource_manager(self):
        """
        Get the `asdf.resource.ResourceManager` instance.  Includes resources from
        registered resource mappings and any mappings added at runtime.

        Returns
        -------
        `asdf.resource.ResourceManager`
        """
        if self._resource_manager is None:
            with self._lock:
                if self._resource_manager is None:
                    self._resource_manager = ResourceManager(self.resource_mappings)
        return self._resource_manager

    @property
    def extensions(self):
        """
        Get the list of registered extensions.

        Returns
        -------
        list of asdf.extension.ExtensionProxy
        """
        if self._extensions is None:
            with self._lock:
                if self._extensions is None:
                    self._extensions = _entry_points.get_extensions()
        return self._extensions

    def add_extension(self, extension):
        """
        Register a new extension.  The new extension will
        take precedence over all previously registered extensions.

        Parameters
        ----------
        extension : asdf.extension.Extension
        """
        with self._lock:
            extension = ExtensionProxy.maybe_wrap(extension)
            self._extensions = [extension] + [e for e in self.extensions if e != extension]

    def remove_extension(self, extension=None, *, package=None):
        """
        Remove a registered extension.

        Parameters
        ----------
        extension : asdf.extension.Extension or str, optional
            An extension instance or URI pattern to remove.
        package : str, optional
            Remove only extensions provided by this package.  If the ``extension``
            argument is omitted, then all extensions from this package will
            be removed.
        """
        if extension is None and package is None:
            msg = "Must specify at least one of extension or package"
            raise ValueError(msg)

        if extension is not None and not isinstance(extension, str):
            extension = ExtensionProxy.maybe_wrap(extension)

        def _remove_condition(e):
            result = True

            if isinstance(extension, str):
                result = result and util.uri_match(extension, e.extension_uri)
            elif isinstance(extension, ExtensionProxy):
                result = result and e == extension

            if package is not None:
                result = result and e.package_name == package

            return result

        with self._lock:
            self._extensions = [e for e in self.extensions if not _remove_condition(e)]

    def reset_extensions(self):
        """
        Reset extensions to the default list registered via entry points.
        """
        with self._lock:
            self._extensions = None

    @property
    def default_version(self):
        """
        Get the default ASDF core schemas version used for
        new files.

        Returns
        -------
        str
        """
        return self._default_version

    @default_version.setter
    def default_version(self, value):
        """
        Set the default ASDF core schemas version used for
        new files.

        Parameters
        ----------
        value : str
        """
        self._default_version = validate_version(value)

    @property
    def io_block_size(self):
        """
        Get the block size used when reading and writing
        files.

        Returns
        -------
        int
            Block size, or -1 to use the filesystem's
            preferred block size.
        """
        return self._io_block_size

    @io_block_size.setter
    def io_block_size(self, value):
        """
        Set the block size used when reading and writing
        files.

        Parameters
        ----------
        value : int
            Block size, or -1 to use the filesystem's
            preferred block size.
        """
        self._io_block_size = value

    @property
    def legacy_fill_schema_defaults(self):
        """
        Get the configuration that controls filling defaults
        from schemas for older ASDF core schemas versions.  If
        `True`, missing default values will be filled from the
        schema when reading files from ASDF core schemas <= 1.5.0.
        Later versions of the core schemas do not support removing
        or filling schema defaults.

        Returns
        -------
        bool
        """
        return self._legacy_fill_schema_defaults

    @legacy_fill_schema_defaults.setter
    def legacy_fill_schema_defaults(self, value):
        """
        Set the flag that controls filling defaults from
        schemas for older ASDF core schemas versions.

        Parameters
        ----------
        value : bool
        """
        self._legacy_fill_schema_defaults = value

    @property
    def array_inline_threshold(self):
        """
        Get the threshold below which arrays are automatically written
        as inline YAML literals instead of binary blocks.  This number
        is compared to number of elements in the array.

        Returns
        -------
        int or None
            Integer threshold, or None to disable automatic selection
            of the array storage type.
        """
        return self._array_inline_threshold

    @array_inline_threshold.setter
    def array_inline_threshold(self, value):
        """
        Set the threshold below which arrays are automatically written
        as inline YAML literals instead of binary blocks.  This number
        is compared to number of elements in the array.

        Parameters
        ----------
        value : int or None
            Integer threshold, or None to disable automatic selection
            of the array storage type.
        """
        self._array_inline_threshold = value

    @property
    def all_array_storage(self):
        """
        Override the array storage type of all blocks
        in the file immediately before writing.  Must be one of the
        following strings or `None`:

        - ``internal``: The default.  The array data will be
          stored in a binary block in the same ASDF file.

        - ``external``: Store the data in a binary block in a
          separate ASDF file.

        - ``inline``: Store the data as YAML inline in the tree.
        """
        return self._all_array_storage

    @all_array_storage.setter
    def all_array_storage(self, value):
        if value not in (None, "internal", "external", "inline"):
            msg = f"Invalid value for all_array_storage: '{value}'"
            raise ValueError(msg)
        self._all_array_storage = value

    @property
    def all_array_compression(self):
        """
        Override the compression type on all binary blocks in the
        file.  Must be one of the following strings, `None` or a
        label supported by a `asdf.extension.Compressor`:

        - ``''`` or `None`: No compression.

        - ``zlib``: Use zlib compression.

        - ``bzp2``: Use bzip2 compression.

        - ``lz4``: Use lz4 compression.

        - ``input``: Use the same compression as in the file read.
          If there is no prior file, acts as None
        """
        return self._all_array_compression

    @all_array_compression.setter
    def all_array_compression(self, value):
        # local to avoid circular import
        from asdf._compression import validate

        self._all_array_compression = validate(value)

    @property
    def all_array_compression_kwargs(self):
        """
        Dictionary of keyword arguments provided to the compressor during
        block compression (or `None` for no keyword arguments)
        """
        return self._all_array_compression_kwargs

    @all_array_compression_kwargs.setter
    def all_array_compression_kwargs(self, value):
        if value is not None and not isinstance(value, collections.abc.Mapping):
            msg = f"Invalid value for all_array_compression_kwargs: '{value}'"
            raise ValueError(msg)
        self._all_array_compression_kwargs = value

    @property
    def default_array_save_base(self):
        """
        Option to control if when saving arrays the base array should be
        saved (so views of the same array will refer to offsets/strides of the
        same block).
        """
        return self._default_array_save_base

    @default_array_save_base.setter
    def default_array_save_base(self, value):
        if not isinstance(value, bool):
            msg = "default_array_save_base must be a bool"
            raise ValueError(msg)
        self._default_array_save_base = value

    @property
    def validate_on_read(self):
        """
        Get configuration that controls schema validation of
        ASDF files on read.

        Returns
        -------
        bool
        """
        return self._validate_on_read

    @validate_on_read.setter
    def validate_on_read(self, value):
        """
        Set the configuration that controls schema validation of
        ASDF files on read.  If `True`, newly opened files will
        be validated.

        Parameters
        ----------
        value : bool
        """
        self._validate_on_read = value

    @property
    def lazy_tree(self):
        """
        Get configuration that controls if ASDF tree contents
        are lazily converted to custom objects or if all custom
        objects are created when the file is opened. See the
        ``lazy_tree`` argument for `asdf.open`.

        Returns
        -------
        bool
        """
        return self._lazy_tree

    @lazy_tree.setter
    def lazy_tree(self, value):
        self._lazy_tree = value

    @property
    def warn_on_failed_conversion(self):
        """
        Get configuration that controls if errors during
        conversion are converted to warnings.

        Enabling this can be helpful when opening old
        files that contain tags that are no longer supported.

        Returns
        -------
        bool
        """
        return self._warn_on_failed_conversion

    @warn_on_failed_conversion.setter
    def warn_on_failed_conversion(self, value):
        self._warn_on_failed_conversion = value

    def __repr__(self):
        return (
            "<AsdfConfig\n"
            f"  array_inline_threshold: {self.array_inline_threshold}\n"
            f"  all_array_storage: {self.all_array_storage}\n"
            f"  all_array_compression: {self.all_array_compression}\n"
            f"  all_array_compression_kwargs: {self.all_array_compression_kwargs}\n"
            f"  default_array_save_base: {self.default_array_save_base}\n"
            f"  default_version: {self.default_version}\n"
            f"  io_block_size: {self.io_block_size}\n"
            f"  legacy_fill_schema_defaults: {self.legacy_fill_schema_defaults}\n"
            f"  validate_on_read: {self.validate_on_read}\n"
            f"  lazy_tree: {self.lazy_tree}\n"
            f"  warn_on_failed_conversion: {self.warn_on_failed_conversion}\n"
            ">"
        )


class _ConfigLocal(threading.local):
    def __init__(self):
        self.config_stack = []


_global_config = AsdfConfig()
_local = _ConfigLocal()


def get_config():
    """
    Get the current config, which may have been altered by
    one or more surrounding calls to `asdf.config_context`.

    Returns
    -------
    asdf.config.AsdfConfig
    """
    if len(_local.config_stack) == 0:
        return _global_config

    return _local.config_stack[-1]


@contextmanager
def config_context():
    """
    Context manager that temporarily overrides asdf configuration.
    The context yields an `asdf.config.AsdfConfig` instance that can be modified
    without affecting code outside of the context.
    """
    base_config = _global_config if len(_local.config_stack) == 0 else _local.config_stack[-1]

    config = copy.copy(base_config)
    _local.config_stack.append(config)

    try:
        yield config
    finally:
        _local.config_stack.pop()
